"""Test filters work as expected."""

from pathlib import Path

import pytest

from castep_outputs import parse_castep_file
from castep_outputs.parsers.castep_file_parser import Filters

_DATA_FOLDER = Path(__file__).parent / "data_files"
_TEST_FILE = _DATA_FOLDER / "test.castep"

@pytest.mark.parametrize("filters,keys", (
    (Filters.CELL, {"initial_cell"}),
    (Filters.PARAMETERS, {"title", "options", "initial_spins", "k-points", "target_stress"}),
    (Filters.POSITION, {"initial_positions"}),
    (Filters.PSPOT, {"pspot_detail"}),
    (Filters.SPECIES_PROPS, {"species_properties"}),
    (Filters.SYMMETRIES, {"symmetries", "constraints"}),
    (Filters.SYS_INFO, {"build_info", "time_started", "warning", "memory_estimate"}),
    (Filters.LOW, {"species_properties", "initial_cell", "initial_positions"}),
    (Filters.MEDIUM, {"species_properties", "title", "options", "initial_cell",
                      "initial_positions", "initial_spins", "k-points", "target_stress"}),
    (Filters.HIGH, {"build_info", "time_started", "warning", "pspot_detail",
                    "species_properties", "title", "options",
                    "initial_cell", "initial_positions", "initial_spins", "k-points",
                    "symmetries", "constraints", "target_stress", "memory_estimate"}),
    (Filters.FULL, {"build_info", "time_started", "warning", "pspot_detail",
                    "species_properties", "title", "options", "initial_cell",
                    "initial_positions", "initial_spins", "k-points", "symmetries",
                    "constraints", "target_stress", "memory_estimate"}),
    (Filters.TESTING, {"pspot_detail", "species_properties", "initial_cell", "initial_positions"}),
))
def test_filter_all(filters, keys):
    with _TEST_FILE.open("r", encoding="utf-8") as file:
        data = parse_castep_file(file, filters=filters)
    assert keys <= data[0].keys()

def test_filter_none():
    with _TEST_FILE.open("r", encoding="utf-8") as file:
        data = parse_castep_file(file, filters=Filters.NONE)
    assert not data


if __name__ == "__main__":
    pytest.main()



"""
Parse the following castep outputs.

- .elf_fmt
- .chdiff_fmt
- .pot_fmt
- .den_fmt
"""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO, TypedDict, cast

from ..utilities.utility import file_or_path, stack_dict, to_type


class ElfFileInfo(TypedDict, total=False):
    """Electron localisation function information."""

    #: :math:`q`-point sample.
    q: list[tuple[int, int, int]]
    #: Spin-split alpha channel localisation
    chi_alpha: tuple[float, ...]
    #: Spin-split beta channel localisation
    chi_beta: tuple[float, ...]
    #: Electron localisation function
    chi: tuple[float, ...]


class ChDiffFileInfo(TypedDict):
    """Charge difference information."""

    #: :math:`q`-point sample.
    q: list[tuple[int, int, int]]
    #: Difference between charge density and superposition of atomic densities.
    charge: tuple[float, ...]


class DenFileInfo(TypedDict):
    """Charge density information."""

    #: :math:`q`-point sample.
    q: list[tuple[int, int, int]]
    #: Charge density at point.
    charge: tuple[float, ...]


class PotFileInfo(TypedDict):
    """Potential information."""

    #: :math:`q`-point sample.
    q: list[tuple[int, int, int]]
    #: Charge density at point.
    pot: tuple[float, ...]


@file_or_path(mode="r")
def _parse_kpt_info(
        inp: TextIO,
) -> dict[str, list[int | float]]:
    """
    Parse standard form of kpt related .*_fmt files.

    Parameters
    ----------
    inp : TextIO
        File to parse.

    Returns
    -------
    dict[str, list[int | float]]
        Parsed data.
    """
    # Skip header
    while "END header" not in (line := inp.readline()):
        pass

    cols = re.search(r'"([^"]+)"', line).group(1)
    prop = tuple(cols.split()[3:]) if len(cols) > 3 else cols[3]

    qdata = defaultdict(list)

    for line in inp:
        if not line.strip():
            continue
        if isinstance(prop, str):
            *qpt, val = line.split()
            qpt = to_type(qpt, int)
            val = to_type(val, float)
            stack_dict(qdata, {"q": qpt, prop: val})
        elif isinstance(prop, tuple):
            words = line.split()
            qpt = to_type(words[0:3], int)
            val = to_type(words[3:], float)
            stack_dict(qdata, {"q": qpt, **dict(zip(prop, val))})

    return qdata


def parse_elf_fmt_file(elf_file: TextIO) -> ElfFileInfo:
    """
    Parse castep .elf_fmt file.

    Parameters
    ----------
    elf_file : ~typing.TextIO
        Open handle to file to parse.

    Returns
    -------
    ElfFileInfo
        Parsed info.
    """
    return cast(ElfFileInfo, _parse_kpt_info(elf_file))


def parse_chdiff_fmt_file(chdiff_file: TextIO) -> ChDiffFileInfo:
    """
    Parse castep .chdiff_fmt file.

    Parameters
    ----------
    chdiff_file : ~typing.TextIO
        Open handle to file to parse.

    Returns
    -------
    ChDiffFileInfo
        Parsed info.
    """
    return cast(ChDiffFileInfo, _parse_kpt_info(chdiff_file))


def parse_pot_fmt_file(pot_file: TextIO) -> PotFileInfo:
    """
    Parse castep .pot_fmt file.

    Parameters
    ----------
    pot_file : ~typing.TextIO
        Open handle to file to parse.

    Returns
    -------
    PotFileInfo
        Parsed info.
    """
    return cast(PotFileInfo, _parse_kpt_info(pot_file))


def parse_den_fmt_file(den_file: TextIO) -> DenFileInfo:
    """
    Parse castep .den_fmt file.

    Parameters
    ----------
    den_file : ~typing.TextIO
        Open handle to file to parse.

    Returns
    -------
    DenFileInfo
        Parsed info.
    """
    return cast(DenFileInfo, _parse_kpt_info(den_file))



"""Module to parse miscellaneous castep files."""

__author__ = "Jacob Wilkins"
__version__ = "0.2.0"

# pylint: disable=unused-import

from .cli.castep_outputs_main import parse_single  # noqa: F401
from .parsers import *  # noqa: F403



"""Parse castep .phonon files."""
from __future__ import annotations

from collections import defaultdict
from typing import TextIO

from ..utilities.datatypes import ComplexThreeVector, QData, StandardHeader, ThreeVector
from ..utilities.filewrapper import Block
from ..utilities.utility import file_or_path, log_factory, to_type
from .parse_utilities import parse_regular_header


class PhononFileQPoint(QData, total=False):
    """Q-point info from a phonon file."""

    #: Direction to next Q-point
    dir: ThreeVector
    #: Eigenvalues of Q-point.
    eigenvalues: list[float]
    #: Eigenvectors of Q-point.
    eigenvectors: list[ComplexThreeVector]


class PhononFileInfo(StandardHeader, total=False):
    """Full phonon file information."""

    #: Number of ions in system.
    ions: int
    #: Number of phonon branches.
    branches: int
    #: Number of wavevectors in calculation.
    wavevectors: int
    #: Per Q-point information.
    qpts: list[PhononFileQPoint]


@file_or_path(mode="r")
def parse_phonon_file(phonon_file: TextIO) -> PhononFileInfo:
    """Parse castep .phonon file.

    Parameters
    ----------
    phonon_file
        A handle to a CASTEP .phonon file.

    Raises
    ------
    ValueError
        Branches count doesn't match expected.

    Returns
    -------
    PhononFileInfo
        Parsed data.
    """
    # pylint: disable=too-many-locals

    logger = log_factory(phonon_file)
    phonon_info: PhononFileInfo = defaultdict(list)

    for line in phonon_file:
        if block := Block.from_re(line, phonon_file, "BEGIN header", "END header"):

            data = parse_regular_header(block)
            phonon_info.update(data)
            del phonon_info[""]
            eigenvectors_endblock = (format(phonon_info["branches"], ">4") +
                                     format(phonon_info["ions"], ">4"))

        elif block := Block.from_re(line, phonon_file, "q-pt", "Phonon Eigenvectors"):

            logger("Found eigenvalue block")

            _, _, posx, posy, posz, weight, *qpoint_dir = line.split()
            qdata: PhononFileQPoint = defaultdict(list)

            qdata["qpt"] = to_type((posx, posy, posz), float)
            qdata["weight"] = float(weight)

            if qpoint_dir:
                qdata["dir"] = to_type(qpoint_dir, float)

            if len(block) - 2 != phonon_info["branches"]:
                raise ValueError("Malformed phonon frequencies. "
                                 f"Expected {phonon_info['branches']} branches, "
                                 f"received {len(block) - 2}")

            block.remove_bounds(1, 1)

            for line in block:
                # Line columns taken from phonon.f90
                eigenvalue, ir_intensity, raman_intensity = (line[8:23],
                                                             line[31:46],
                                                             line[55:70])
                qdata["eigenvalues"].append(float(eigenvalue))
                if ir_intensity.strip():
                    qdata["ir_intensity"].append(float(ir_intensity))
                if raman_intensity.strip():
                    qdata["raman_intensity"].append(float(raman_intensity))

        elif block := Block.from_re(line, phonon_file, "Mode Ion", eigenvectors_endblock):

            logger("Found eigenvector block")

            expected_n_eigenvec = phonon_info["branches"] * phonon_info["ions"]
            if len(block) - 1 != expected_n_eigenvec:
                raise ValueError("Malformed eigenvectors. "
                                 f"Expected {expected_n_eigenvec} branches, "
                                 f"received {len(block) - 1}")

            block.remove_bounds(1, 0)

            for line in block:
                _, _, *eigenvectors = line.split()
                eigenvectors = to_type(eigenvectors, float)
                eigenvectors = map(complex, eigenvectors[::2], eigenvectors[1::2])
                qdata["eigenvectors"].append(tuple(eigenvectors))

            phonon_info["qpts"].append(qdata)

    return phonon_info



"""Parse castep .xrd_sf files."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO, TypedDict

from ..utilities import castep_res as REs
from ..utilities.castep_res import labelled_floats
from ..utilities.datatypes import ThreeVector
from ..utilities.utility import file_or_path, stack_dict, to_type


class XRDSFFileInfo(TypedDict):
    """X-Ray Structure Factor computation."""

    #: Structure factor contribution due to the all-electron valence
    #: augmentation charge.
    f_aug: list[complex]
    #: Indepedent atom model structure factor; electron density formed
    #: from superposition of atomic electron densities obtained by
    #: numerically solving the Koelling-Harmon equation.
    f_iam: list[complex]
    #: DFT structure factor with all-electron augmentation charge.
    f_paw: list[complex]
    #: DFT structure factor with pseudised (inaccurate) augmentation charge
    f_pp: list[complex]
    #: Structure factor contribution to `f_paw` due to the (frozen) core electrons.
    f_core: list[complex]
    #: Structure factor contribution to `f_paw` due to the soft valence
    #: charge (where augmentation charge contribute is removed).
    f_soft: list[complex]
    #: Reciprocal lattice directions of structure factor calculation.
    qvec: list[ThreeVector]


@file_or_path(mode="r")
def parse_xrd_sf_file(xrd_sf_file: TextIO) -> XRDSFFileInfo:
    """
    Parse castep .xrd_sf file.

    Parameters
    ----------
    xrd_sf_file
        Open handle to file to parse.

    Returns
    -------
    XRDSFFileInfo
        Parsed info.
    """
    # Get headers from first line
    headers = xrd_sf_file.readline().split()[3:]
    # Turn Re(x) into x_re
    headers = [f"{head[3:-1]}_{head[0:2]}".lower() for head in headers]
    headers_wo = {head[:-3] for head in headers}

    xrd_re = re.compile(rf"(?P<qvec>(?:\s*{REs.INTNUMBER_RE}){{3}})" +
                        labelled_floats(headers))

    xrd: XRDSFFileInfo = defaultdict(list)
    for line in xrd_sf_file:
        match = xrd_re.match(line).groupdict()
        accum = {head: complex(float(match[f"{head}_re"]),
                               float(match[f"{head}_im"]))
                 for head in headers_wo}
        accum["qvec"] = to_type(match["qvec"].split(), int)
        stack_dict(xrd, accum)

    return xrd



"""Parse castep .ts files."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO, TypedDict

from ..utilities.castep_res import ATOMIC_DATA_TAG, TAG_RE, get_numbers, labelled_floats
from ..utilities.constants import FST_D, TAG_ALIASES, TS_TYPES
from ..utilities.datatypes import ThreeByThreeMatrix
from ..utilities.filewrapper import Block
from ..utilities.utility import add_aliases, atreg_to_index, file_or_path, to_type


class TSStructInfo(TypedDict, total=False):
    """
    Structure info of the intermediate states of a TSS.

    Notes
    -----
    Also contains atom index keys of {AtomIndex: dict}
    """

    #: Total Energy and Enthalpy of system in Ha.
    energy: tuple[float, float]
    #: Alias of :attr:`energy`.
    E: tuple[float, float]
    #: Cell vectors of system in Bohr.
    lattice_vectors: ThreeByThreeMatrix
    #: Alias of :attr:`lattice_vectors`.
    h: ThreeByThreeMatrix


class TSFileInfo(TypedDict):
    """Transition state search file info."""

    #: Calculation is TSS confirmation.
    confirmation: bool
    #: Reagent info.
    reagent: TSStructInfo
    #: Product info.
    product: TSStructInfo
    #: Test info.
    test: TSStructInfo


@file_or_path(mode="r")
def parse_ts_file(ts_file: TextIO) -> TSFileInfo:
    """
    Parse castep .ts file.

    Parameters
    ----------
    ts_file
        Open handle to file to parse.

    Returns
    -------
    TSFileInfo
        Parsed info.
    """
    accum = defaultdict(list)

    for line in ts_file:
        if "TSConfirmation" in line:
            accum["confirmation"] = True

        elif block := Block.from_re(line, ts_file, "(REA|PRO|TST)", r"^\s*$", eof_possible=True):
            curr = defaultdict(list)
            match = re.match(r"\s*(?P<type>REA|PRO|TST)\s*\d+\s*" +
                             labelled_floats(("reaction_coordinate",)), line)
            key = TS_TYPES[match["type"]]
            curr["reaction_coordinate"] = to_type(match["reaction_coordinate"], float)

            for blk_line in block:
                if match := ATOMIC_DATA_TAG.search(blk_line):
                    ion = atreg_to_index(match)
                    if ion not in curr:
                        curr[ion] = {}
                    curr[ion][match.group("tag")] = to_type([*(match.group(d)
                                                               for d in FST_D)], float)
                    add_aliases(curr[ion], TAG_ALIASES)

                elif match := TAG_RE.search(blk_line):
                    curr[match.group("tag")].append([*to_type(get_numbers(blk_line), float)])

            add_aliases(curr, TAG_ALIASES)
            accum[key].append(curr)

    return accum



"""Parse castep .magres files."""

from __future__ import annotations

from collections import defaultdict
from typing import Literal, TextIO, TypedDict

import castep_outputs.utilities.castep_res as REs

from ..utilities.castep_res import get_numbers
from ..utilities.datatypes import AtomIndex, ThreeByThreeMatrix, ThreeVector
from ..utilities.filewrapper import Block
from ..utilities.utility import (
    add_aliases,
    determine_type,
    file_or_path,
    normalise_key,
    to_type,
)


class AtomsInfo(TypedDict):
    """Per atom magres information."""

    #: Susceptability anisotropy.
    anisotropy: float
    #: Isotropic susceptability.
    isotropic: float
    #: Asymmetry ratio.
    asymmetry: float
    #: Atom coordinates.
    coords: ThreeVector
    #: Eigenvalues in :math:`\sigma_{\alpha\alpha}`.
    eigenvalue: ThreeVector
    #: Eigenvectors in :math:`\sigma_{\alpha\alpha}`.
    eigenvector: ThreeByThreeMatrix


class UnitsInfo(TypedDict):
    """Information on magres units."""

    #: Position units.
    atom: str
    #: Lattice length units.
    lattice: str
    #: Magnetic reponse units.
    ms: str


class MagresInfo(TypedDict):
    """NMR Magnetic reponse information."""

    #: Atom coordinates
    atoms: dict[AtomIndex, ThreeVector]
    calc_code: Literal["CASTEP"]
    calc_code_hgversion: str
    calc_code_platform: str
    calc_code_version: str
    calc_cutoffenergy: str
    calc_kpoint_mp_grid: str
    calc_kpoint_mp_offset: str
    calc_name: str
    calc_pspot: str
    calc_xcfunctional: str
    #: Crystal lattice.
    lattice: tuple[float, float, float, float, float, float, float, float, float]
    #: Magnetic susceptibility tensor.
    ms: dict[AtomIndex, list[float]]
    units: UnitsInfo


@file_or_path(mode="r")
def parse_magres_file(magres_file: TextIO) -> MagresInfo:
    """
    Parse castep .magres file.

    Parameters
    ----------
    magres_file
        Open handle to file to parse.

    Returns
    -------
    MagresInfo
        Parsed info.
    """
    # pylint: disable=too-many-branches

    accum = defaultdict(dict)

    for line in magres_file:
        if block := Block.from_re(line, magres_file, r"^\s*\[\w+\]\s*$", r"^\s*\[/\w+\]\s*$"):
            block_name = next(block).strip().strip("[").strip("]")

            if block_name == "calculation":
                accum["calc"] = _process_calculation_block(block)
                version = int(accum["calc"]["code_version"].split(".")[0])
            elif block_name == "atoms":
                accum["atoms"] = _process_atoms_block(block)
            elif block_name == "magres":
                accum["magres"] = _process_magres_block(block, version)
            elif block_name == "magres_old":
                (accum["total_shielding"], accum["atom_info"]) = _process_magres_old_block(
                    block, accum,
                )

    return accum


def _process_calculation_block(block: Block) -> dict[str, str]:
    """
    Process a calculation parameters block.

    Parameters
    ----------
    block : Block
        Block to process

    Returns
    -------
    dict[str, str]
        Processed data.
    """
    calc = {}
    for line in block:
        if line.startswith("["):
            continue
        key, *val = line.strip().split(maxsplit=1)

        if val:
            # Strip `calc_`
            calc[key[5:]] = val[0]
    return calc


def _process_atoms_block(block: Block) -> dict[AtomIndex, ThreeVector]:
    """
    Process the atoms block.

    Parameters
    ----------
    block : Block
        Block to process

    Returns
    -------
    dict[str, str]
        Processed data.
    """
    accum = {"units": {}, "coords": {}}
    for line in block:
        if line.startswith("lattice"):
            key, *val = line.split()
            accum[key] = to_type(val, float)

        elif line.startswith("atom"):
            key, _, spec, ind, *pos = line.split()
            accum["coords"][spec, int(ind)] = to_type(pos, float)

        elif line.startswith("units"):
            _, key, val = line.split()
            accum["units"][key] = val
    return accum


def _process_magres_block(block: Block, version: int) -> dict[str, str | ThreeByThreeMatrix]:
    """
    Process the main magres block.

    Parameters
    ----------
    block : Block
        Block to process

    Returns
    -------
    dict[str, str]
        Processed data.
    """
    accum = defaultdict(dict)

    munge_fix = 3 if version < 22 else 5

    for line in block:
        if line.startswith("units"):
            _, key, val = line.split()
            accum["units"][key] = val

        elif (words := line.split())[0] in {"ms", "efg", "efg_local", "efg_nonlocal"}:
            key, spec, ind, *val = words

            if determine_type(ind) is float:  # Have a munged spec-ind
                val.insert(0, ind)
                spec, ind = spec[:-munge_fix], spec[-munge_fix:]

            accum[key][spec, int(ind)] = to_type(val, float)

        elif words[0].startswith("isc"):  # ISC props explicitly have spaces!
            key, speca, inda, specb, indb, *val = words

            accum[key][(speca, int(inda)), (specb, int(indb))] = to_type(val, float)

    add_aliases(
        accum,
        {
            "ms": "magnetic_resonance_shielding",
            "efg": "electric_field_gradient",
            "efg_local": "local_electric_field_gradient",
            "efg_nonlocal": "nonlocal_electric_field_gradient",
            "isc_fc": "j_coupling_fc",
            "isc_orbital_p": "j_coupling_orbital_p",
            "isc_orbital_d": "j_coupling_orbital_d",
            "isc_spin": "j_coupling_spin",
            "isc": "j_coupling_k_total",
        },
    )

    return accum


def _process_magres_old_block(
    block: Block,
    accum: dict,
) -> tuple[ThreeByThreeMatrix, dict[AtomIndex, AtomsInfo]]:
    """
    Process the magres_old block.

    Parameters
    ----------
    block : Block
        Block to process.

    Returns
    -------
    ThreeByThreeMatrix
        Total shielding tensor.
    dict[AtomIndex, AtomsInfo]
        Processed atomic data.
    """
    atom_info = {atreg: defaultdict(list) for atreg in accum["atoms"]["coords"]}
    for atreg, pos in accum["atoms"]["coords"].items():
        atom_info[atreg]["coords"] = pos

    total_shielding = None

    for line in block:
        if sub_blk := Block.from_re(line, block, "TOTAL Shielding Tensor", REs.EMPTY, n_end=2):
            total_shielding = [
                to_type(numbers, float)
                for sub_line in sub_blk
                if (numbers := get_numbers(sub_line))
            ]
        elif Block.from_re(line, block, "^=+$", "^=+$"):
            pass
        elif line.strip() and "[" not in line:
            spec, ind, key, *val = line.split()
            atreg = spec, int(ind)
            if key == "Eigenvalue":
                atom_info[atreg]["eigvenvalue"].append(float(val[1]))
            elif key == "Eigenvector":
                atom_info[atreg]["eigvenvector"].append(to_type(val[1:], float))
            elif ":" in key:
                atom_info[atreg][normalise_key(key)] = float(val[0])

    return total_shielding, atom_info



"""Parse castep .epme files."""

from __future__ import annotations

from typing import TextIO, TypedDict

from ..utilities.castep_res import get_numbers
from ..utilities.datatypes import ThreeVector
from ..utilities.utility import file_or_path, log_factory, to_type


class EPMEData(TypedDict):
    """Single line of EPME data."""

    # Initial k-point
    kpt_i: ThreeVector
    # Final k-point
    kpt_f: ThreeVector
    # Branch index.
    ph_index: int
    # Initial band.
    band_i: int
    # Final band.
    band_f: int
    # Phonon frequency.
    freq: float
    # Initial energy.
    e_i: float
    # Final Energy.
    e_f: float
    # Velocities.
    velocity: tuple[ThreeVector, ThreeVector]
    # Matrix element.
    e_p_matrix: complex


class EPMEFileInfo(TypedDict):
    """Full EPME file data."""

    # File version.
    version: int
    # Unit cell volume.
    volume: float
    # Fermi energy.
    fermi_energy: float
    # Number of bands.
    n_bands: int
    # Number of phonon branches.
    n_branches: int
    # EPME data.
    data: list[EPMEData]


def _parse_epme_header(epme_file: TextIO) -> EPMEFileInfo:
    """Read header and prepare EPME output dictionary.

    Parameters
    ----------
    epme_file : TextIO
        File to parse.

    Returns
    -------
    EPMEFileInfo
        Prepared EPME data with header.
    """
    epme_data: EPMEFileInfo = {}
    next(epme_file)  # Title
    epme_data["version"] = int(next(epme_file).rsplit(maxsplit=1)[1])
    next(epme_file)  # Units
    epme_data["volume"] = float(next(epme_file).rsplit(maxsplit=1)[1])
    epme_data["fermi_energy"] = float(next(epme_file).rsplit(maxsplit=1)[1])
    epme_data["n_bands"] = int(next(epme_file).rsplit(maxsplit=1)[1])
    epme_data["n_branches"] = int(next(epme_file).rsplit(maxsplit=1)[1])
    epme_data["data"] = []

    return epme_data


@file_or_path(mode="r")
def parse_epme_file(epme_file: TextIO) -> EPMEFileInfo:
    """Parse castep .epme file.

    Parameters
    ----------
    epme_file
        A handle to a CASTEP .epme file.

    Returns
    -------
    EPMEFileInfo
        Parsed data.
    """
    logger = log_factory(epme_file)

    epme_data: EPMEFileInfo = _parse_epme_header(epme_file)
    kpt_i, kpt_f = None, None

    for line in epme_file:
        if line.startswith("Electron-Phonon coupling"):
            kpts = to_type(get_numbers(line), float)
            kpt_i, kpt_f = kpts[:3], kpts[3:]
            logger("Found k-point pair: %s -> %s", kpt_i, kpt_f)
            next(epme_file)  # Skip header
            continue
        data = line.split()
        curr = {"kpt_i": kpt_i, "kpt_f": kpt_f}
        curr["ph_index"], curr["band_i"], curr["band_f"] = to_type(data[:3], int)
        curr["e_i"], curr["e_f"] = to_type(data[3:5], float)
        curr["velocity"] = to_type(data[5:8], float), to_type(data[8:11], float)
        curr["e_p_matrix"] = complex(*to_type(data[11:13], float))
        epme_data["data"].append(curr)

    return epme_data



"""Module containing all regexes."""

from __future__ import annotations

import itertools
import re
from collections.abc import Sequence
from typing import Union

from .constants import FST_D, MINIMISERS, SHELLS, SND_D

#: Valid input where patterns are wanted.
Pattern = Union[str, re.Pattern]


def get_numbers(line: str) -> list[str]:
    """
    Get all numbers in a string as a list.

    Parameters
    ----------
    line
        String to search for valid numbers.

    Returns
    -------
    list[str]
        All numbers found in `line`.

    Examples
    --------
    >>> get_numbers("Hello 13, my name is 3.142")
    ['13', '3.142']

    See Also
    --------
    NUMBER_RE : RegEx matching numbers.
    """
    return NUMBER_RE.findall(line)


def labelled_floats(
    labels: Sequence[str],
    counts: Sequence[int | str | None] = (None,),
    sep: str = r"\s+?",
    suffix: str = "",
) -> str:
    """
    Construct a regex for extracting floats with assigned labels.

    Parameters
    ----------
    labels
        Iterable of labels to label each group.
    counts
        Iterable of counts to group into each label (count must not exceed that of labels).
    sep
        Separator between floats.
    suffix
        Suffix following each float.

    Returns
    -------
    str
        Regular expression to match floats with given labels.

    Raises
    ------
    NotImplementedError
        If `suffix` and `counts` provided.

    Examples
    --------
    .. code-block:: python

       # Capture 3 space-separated floats with labels "x", "y", "z"
       labelled_floats(("x", "y", "z"))
       # Capture 2 colon-separated floats with labels "u", "v"
       labelled_floats(("u", "v"), sep=":")
       # Capture 3 space-separated floats under the label "val"
       labelled_floats(("val",), counts=(3,))
    """
    if suffix and any(cnt for cnt in counts):
        raise NotImplementedError("Suffix and counts not currently supported")

    outstr = ""
    for label, cnt in itertools.zip_longest(labels, counts):
        if cnt:
            outstr += f"(?:(?P<{label}>(?:{sep}{NUMBER_RE.pattern}{suffix}){{{cnt}}}))"
        else:
            outstr += f"(?:{sep}(?P<{label}>{NUMBER_RE.pattern}){suffix})"

    return outstr


def gen_table_re(
    content: str,
    border: str = r"\s*",
    *,
    pre: str = "",
    post: str = "",
    whole_line: bool = True,
) -> str:
    r"""
    Construct a regex for matching table headers with given borders.

    Parameters
    ----------
    content
        RegEx capturing the content of the table.
    border
        RegEx capturing the characters the make up the border.
    pre
        RegEx matching content outside and before the table.
    post
        RegEx matching content outside and after the table.
    whole_line
        Whether the RegExes are an exact match for the whole line.

    Returns
    -------
    str
        RegEx pattern to match line(s).

    Examples
    --------
    RegEx for a line of "?"s

    .. code-block:: python

       # No content, with any number of questionmarks.
       gen_table_re("", r"\?+")

    RegEx for a separator line in a table which resembles:

    .. code-block::

       +-----------------------------------+

    .. code-block:: python

       gen_table_re("-+", "\+")

    """
    tab_re = rf"\s*{border}\s*{content}\s*{border}\s*" if content else rf"\s*{border}\s*"

    tab_re = rf"\s*{pre}{tab_re}{post}\s*"

    if whole_line:
        tab_re = f"^{tab_re}$"

    return tab_re


def get_atom_parts(spec: str) -> dict[str, str]:
    """
    Get components of an atom name (species, tag, label) from a spec string.

    Parameters
    ----------
    spec
        String to split up.

    Returns
    -------
    dict[str, str]
        Dictionary detailing separate parts of atom name.

    Examples
    --------
    >>> get_atom_parts("Ar")
    {'species': 'Ar'}
    >>> get_atom_parts("Ar:tag")
    {'species': 'Ar', 'tag': 'tag'}
    >>> get_atom_parts("Ar [label]")
    {'species': 'Ar', 'label': 'label'}
    >>> get_atom_parts("Ar:tag[label]")
    {'species': 'Ar', 'tag': 'tag', 'label': 'label'}
    """
    return {
        key: val.strip() for key, val in ATOM_NAME_GRP_RE.match(spec).groupdict().items() if val
    }


# --- RegExes
# Regexps to recognise numbers

#: Floating point number.
#:
#: :meta hide-value:
FNUMBER_RE = r"(?:[+-]?(?:\d*\.\d+|\d+\.\d*))"

#: Integer number.
#:
#: :meta hide-value:
INTNUMBER_RE = r"(?:[+-]?(?<!\.)\d+(?!\.))"

#: Float or int with exponent.
#:
#: :meta hide-value:
EXPNUMBER_RE = rf"(?:(?:{FNUMBER_RE}|{INTNUMBER_RE})[Ee][+-]?\d{{1,3}})"

#: Exponent or bare floating point.
#:
#: :meta hide-value:
EXPFNUMBER_RE = f"(?:{EXPNUMBER_RE}|{FNUMBER_RE})"

#: Integer ratio.
#:
#: :meta hide-value:
RATIONAL_RE = rf"\b{INTNUMBER_RE}/{INTNUMBER_RE}\b"

#: Generalised number.
#:
#: :meta hide-value:
NUMBER_RE = re.compile(rf"(?:{EXPNUMBER_RE}|{FNUMBER_RE}|{INTNUMBER_RE})")

#: Floating point, integer or ratio.
#:
#: :meta hide-value:
FLOAT_RAT_RE = re.compile(rf"(?:{RATIONAL_RE}|{EXPFNUMBER_RE}|{INTNUMBER_RE})")

#: Three vector.
#:
#: :meta hide-value:
THREEVEC_RE = labelled_floats(("val",), counts=(3,))

# Regexp to identify extended chemical species
#: Element name.
#:
#: :meta hide-value:
SPECIES_RE = r"[A-Z][a-z]{0,2}"
#: CASTEP atom name with optional tag and label.
#:
#: :meta hide-value:
ATOM_NAME_RE = rf"\b{SPECIES_RE}(?::\w+)?\b(?:\s*\[[^\]]+\])?"
#: Labelled CASTEP atom name.
#:
#: :meta hide-value:
ATOM_NAME_GRP_RE = re.compile(
    rf"(?P<species>{SPECIES_RE})(?::(?P<tag>\w+))?\b(?:\s*\[(?P<label>[^\]]*)\])?",
)


# Unless we have *VERY* exotic electron shells

#: RegEx for atomic shell labels. E.g. 4d4.
#:
#: :meta hide-value:
SHELL_RE = rf"\d[{''.join(SHELLS)}]\d{{0,2}}"

#: RegEx for CASTEP table tags as in e.g. .md/.geom.
#:
#: :meta hide-value:
TAG_RE = re.compile(r"<--\s*(?P<tag>\w+)")

#: Empty line RegEx
#:
#: :meta hide-value:
EMPTY = r"^\s*$"

#: CASTEP Atom RegEx with optional index.
#:
#: :meta hide-value:
ATOM_RE = rf"(?P<spec>{ATOM_NAME_RE})\s+(?P<index>\d+)"


# Atom reference with 3-vector
#: CASTEP atom followed by 3D vector.
#:
#: :meta hide-value:
ATOMIC_DATA_3VEC = re.compile(ATOM_RE + labelled_floats(FST_D))
ATOMIC_DATA_FORCE = re.compile(ATOM_RE + labelled_floats(FST_D, suffix=r"(?:\s*\([^)]+\))?"))
ATOMIC_DATA_TAG = re.compile(rf"\s*{ATOMIC_DATA_3VEC.pattern}\s*{TAG_RE.pattern}")

# Labelled positions
LABELLED_POS_RE = re.compile(ATOMIC_DATA_3VEC.pattern + r"\s{5}(?P<label>\S{0,8})")

# Colon separated key-val
COLON_SEP_RE = re.compile(r"\b(?P<key>\S+)\s*:\s*(?P<val>\S+)\b")

# VCA atoms
MIXTURE_LINE_1_RE = re.compile(
    rf"""
(?P<index>\d+)\s+
{labelled_floats(FST_D)}\s+
(?P<spec>{ATOM_NAME_RE})\s+
{labelled_floats(("weight",))}""",
    re.VERBOSE,
)
MIXTURE_LINE_2_RE = re.compile(rf"(?P<spec>{ATOM_NAME_RE})\s+{labelled_floats(('weight',))}")


# SCF Loop
SCF_LOOP_RE = re.compile(
    r"\s*(?:Initial|\d+)\s*"
    rf"{labelled_floats(('energy', 'fermi_energy', 'energy_gain'))}?\s*"
    f"{labelled_floats(('time',))}",
)

# Spin density
INTEGRATED_SPIN_DENSITY_RE = re.compile(
    r"(?P<vec>2\*)?Integrated \|?Spin Density\|?[^=]+=\s*"
    rf"(?P<val>{EXPFNUMBER_RE}\s*"
    rf"(?(vec)(?:{EXPFNUMBER_RE}\s*){{2}}))",
)

# PS Energy
PS_SHELL_RE = re.compile(
    rf"\s*Pseudo atomic calculation performed for (?P<spec>{SPECIES_RE})(\s+{SHELL_RE})+",
)

# PS Projector
PROJ_TYPES = "[UNGHLP]"
PSPOT_SHELL_RE = rf"(?:{SHELL_RE}{FNUMBER_RE}?)"
PSPOT_PROJ_GROUPS = ("orbital", "shell", "type", "de", "beta_delta")
PSPOT_PROJ_RE = re.compile(
    rf"""
    (\d)                            # Orbital
    (\d)                            # Shell type
    ({PROJ_TYPES}*)                 # Proj type
    ((?:[\+-=]{FNUMBER_RE})?)       # DE_use
    ((?:@{FNUMBER_RE})?)            # beta_delta
""",
    re.VERBOSE,
)

PSPOT_REFERENCE_STRUC_RE = re.compile(
    rf"""
    ^\s*\|\s*
    (?P<orb>{SHELL_RE}(?:/\d+)?)\s*
    {labelled_floats(("occupation", "energy"))}
    \s*\|\s*$
    """,
    re.VERBOSE,
)

PSPOT_DEF_RE = re.compile(
    rf"""
    ^\s*\|\s*
    (?P<beta>\d+|loc)\s*
    (?P<l>\d+)\s*
    (?P<j>\d+)?\s*
    {labelled_floats(("e", "Rc"))}\s*
    (?P<scheme>\w+)\s*
    (?P<norm>\d+)
    \s*\|\s*$
    """,
    re.VERBOSE,
)

# PSPot String
PSPOT_RE = re.compile(
    rf"""^
    (?:{labelled_floats(("local_energy",), sep="")}=)?
    (?P<local_channel>\d+)
    (?P<poly_fit>-)?\|
    {labelled_floats(("core_radius",), sep="")}\|
    (?:{labelled_floats(("beta_radius",), sep="")}\|)?
    (?(beta_radius){labelled_floats(("r_inner",), sep="")}\| |)
    (?:{labelled_floats(("coarse",), sep="")}\|)
    (?:{labelled_floats(("medium",), sep="")}\|)
    (?:{labelled_floats(("fine",), sep="")}\|)
    (?P<proj>{PSPOT_PROJ_RE.pattern}(?::{PSPOT_PROJ_RE.pattern})*)
    (?:\{{(?P<shell_swp>{SHELL_RE}(?:,{SHELL_RE})*)\}})?
    (?:\((?P<opt>[^)]+)\))?
    (?P<debug>\#)?
    (?P<print>\[(?P<shell_swp_end> {PSPOT_SHELL_RE}(?:,{PSPOT_SHELL_RE})* )?\])?
    $
""",
    re.VERBOSE,
)

# Forces block
FORCES_BLOCK_RE = re.compile(gen_table_re("([a-zA-Z ]*)Forces", r"\*+"), re.IGNORECASE)
# Stresses block
STRESSES_BLOCK_RE = re.compile(gen_table_re("([a-zA-Z ]*)Stress Tensor", r"\*+"), re.IGNORECASE)

# Bonds
BOND_RE = re.compile(
    rf"""\s*
    (?P<spec1>{ATOM_NAME_RE})\s*(?P<ind1>\d+)\s*
    --\s*
    (?P<spec2>{ATOM_NAME_RE})\s*(?P<ind2>\d+)\s*
    {labelled_floats(("population", "spin", "length"), counts=(None, "0,1", None))}
    """,
    re.VERBOSE,
)

# Pair pot
PAIR_POT_RES = {
    "two_body_one_spec": re.compile(
        rf"^(?P<tag>\w+)?\s*\*\s*(?P<spec>{ATOM_NAME_RE})\s*\*\s*$",
    ),
    "two_body_spec": re.compile(
        rf"(?P<spec1>{ATOM_NAME_RE})\s*-\s*"
        rf"(?P<spec2>{ATOM_NAME_RE})",
    ),
    "two_body_val": re.compile(
        rf"""
            (?P<tag>\w+)?\s*\*\s*
            (?P<label>\w+)\s*
            {labelled_floats(("params",), counts=("1,4",))}\s*
            [\w^/*]+\s* \* \s*
            <--\s*(?P<type>\w+)
            """,
        re.ASCII | re.VERBOSE,
    ),
    "three_body_spec": re.compile(
        rf"""
        ^(?P<tag>\w+)?\s*\*\s*
        (?P<spec>(?:{ATOM_NAME_RE}\s*){{3}})
        \s*\*\s*$""",
        re.VERBOSE,
    ),
    "three_body_val": re.compile(
        rf"""
        ^(?P<tag>\w+)?\s*\*\s*
        (?P<label>\w+)\s*
        {labelled_floats(("params",))}\s*
        [\w^/*]+\s* \* \s*
        <--\s*(?P<type>\w+)
        """,
        re.VERBOSE,
    ),
}

# Orbital population
ORBITAL_POPN_RE = re.compile(
    rf"\s*{ATOM_RE}\s*(?P<orb>[SPDF][xyz]?)"
    rf"\s*{labelled_floats(('charge',))}",
)

# Regexp to identify phonon block in .castep file
PHONON_RE = re.compile(
    rf"""
    \s+\+\s+
    q-pt=\s*{INTNUMBER_RE}\s+
    \({labelled_floats(("qpt",), counts=(3,))}\)
    \s+
    ({FNUMBER_RE})\s+\+
    """,
    re.VERBOSE,
)

PROCESS_PHONON_RE = re.compile(
    rf"""\s+\+\s+
    (?P<N>\d+)\s+
    (?P<frequency>{FNUMBER_RE})\s*
    (?P<irrep>[a-zA-V])?\s*
    (?P<ir_intensity>{FNUMBER_RE})?\s*
    (?P<ir_active>[YN])?\s*
    (?P<raman_intensity>{FNUMBER_RE})?\s*
    (?P<raman_active>[YN])?\s*\+""",
    re.VERBOSE,
)

TDDFT_RE = re.compile(
    rf"""\s*\+\s*
    {INTNUMBER_RE}
    {labelled_floats(("energy", "error"))}
    \s*(?P<type>\w+)
    \s*\+TDDFT""",
    re.VERBOSE,
)

TDDFT_STATE_RE = re.compile(
    rf"""
    \s*(?P<state>\d+)
    \s*(?P<occ>\d+)\s*-->
    \s*(?P<unocc>\d+)
    \s*{labelled_floats(("overlap",))}
    """,
    re.VERBOSE,
)

BS_RE = re.compile(
    rf"""
    Spin=\s*(?P<spin>{INTNUMBER_RE})\s*
    kpt=\s*{INTNUMBER_RE}\s*
    \({labelled_floats(("kx", "ky", "kz"))}\)\s*
    kpt-group=\s*(?P<kpgrp>{INTNUMBER_RE})
    """,
    re.VERBOSE,
)

THERMODYNAMICS_DATA_RE = re.compile(labelled_floats(("t", "e", "f", "s", "cv")))
ATOMIC_DISP_RE = re.compile(
    labelled_floats(("temperature",))
    + r"\s*"
    + ATOM_RE
    + r"\s*"
    + labelled_floats(("displacement",), counts=(6,)),
)

MINIMISERS_RE = f"(?:{'|'.join(x.upper() for x in MINIMISERS)})"
GEOMOPT_MIN_TABLE_RE = re.compile(
    r"\s*\|\s* (?P<step>[^|]+)"
    + labelled_floats(("lambda", "fdelta", "enthalpy"), sep=r"\s*\|\s*")
    + r"\s* \|",
    re.VERBOSE,
)

GEOMOPT_TABLE_RE = re.compile(
    r"\s*\|\s* (?P<parameter>\S+)"
    + labelled_floats(("value", "tolerance"), sep=r"\s*\|\s*")
    + r"\s*\|\s* \S+ (?#Units) \s*\|\s* (?P<converged>No|Yes) \s*\|",
    re.VERBOSE,
)


# Regexp to identify Mulliken ppoulation analysis line
POPN_RE = re.compile(
    rf"\s*{ATOM_RE}\s*(?P<spin_sep>up:)?"
    + labelled_floats((*SHELLS, "total", "charge", "spin"))
    + "?",  # Spin is optional
)

POPN_RE_DN = re.compile(
    r"\s+\d+\s*dn:" + labelled_floats((*SHELLS, "total")),
)

# Regexp for born charges
BORN_RE = re.compile(rf"\s+{ATOM_RE}(?P<charges>(?:\s*{FNUMBER_RE}){{3}})(?:\s*ID=(?P<label>\S+))?")

# MagRes REs

_JC_LABELS = ("fermi_contact", "spin_dipolar", "paramagnetic", "diamagnetic", "total")

MAGRES_RE = (
    # "Chemical Shielding Tensor" 0
    re.compile(
        rf"\s*\|\s*{ATOM_RE}{labelled_floats(('cs_iso', 'cs_aniso'))}\s*"
        rf"(?P<asym>{FNUMBER_RE}|N/A)\s*\|\s*",
    ),
    # "Chemical Shielding and Electric Field Gradient Tensor" 1
    re.compile(
        rf"\s*\|\s*{ATOM_RE}{labelled_floats(('cs_iso', 'cs_aniso'))}\s*"
        rf"(?P<asym>{FNUMBER_RE}|N/A)"
        rf"{labelled_floats(('cq', 'eta'))}\s*\|\s*",
    ),
    # "Electric Field Gradient Tensor" 2
    re.compile(
        rf"\s*\|\s*{ATOM_RE}{labelled_floats(('cq',))}\s*"
        rf"(?P<asym>{FNUMBER_RE}|N/A)\s*\|\s*",
    ),
    # "(?:I|Ani)sotropic J-coupling" 3
    re.compile(
        rf"\s*\|\**\s*{ATOM_RE}"
        rf"{labelled_floats(_JC_LABELS)}\s*\|\s*",
    ),
    # "Hyperfine Tensor" 4
    re.compile(rf"\s*\|\s*{ATOM_RE}{labelled_floats(('hf_iso',))}\s*\|\s*"),
)

# MagRes Tasks
MAGRES_TASK = (
    "Chemical Shielding",
    "Chemical Shielding and Electric Field Gradient",
    "Electric Field Gradient",
    "(An)Isotropic J-coupling",
    "Hyperfine",
)

# Regexp to identify block in .phonon or .phonon_dos file
FRACCOORDS_RE = re.compile(
    rf"\s*(?P<index>{INTNUMBER_RE}){labelled_floats(('u', 'v', 'w'))}"
    rf"\s*(?P<spec>{SPECIES_RE}){labelled_floats(('mass',))}",
)

PHONON_PHONON_RE = re.compile(
    rf"""
    \s+q-pt=\s*{INTNUMBER_RE}\s*
    {labelled_floats(("qpt", "pth"), counts=(3, 1))}
    """,
    re.VERBOSE,
)

PROCESS_PHONON_PHONON_RE = re.compile(labelled_floats(("n", "f", "Grad_qf")))


# Regexp to identify Fermi energies in .bands file
BANDS_FERMI_RE = re.compile(
    r"Fermi energ(ies|y) \(in atomic units\)\s*" + labelled_floats(("a", "b")),
)

# Regexp to identify eigenvalue block in .bands file
# BANDS_EIGENS_RE =
# rf"K-point\s+(\d+)\s*(\s*{FNUMBER_RE})\s*({FNUMBER_RE})\s*({FNUMBER_RE})\s*({FNUMBER_RE})"

DEVEL_CODE_VAL_RE = r"[A-Za-z0-9_]+[:=]\S+"
DEVEL_CODE_BLOCK_RE = rf"([A-Za-z0-9_]+):(?:\s*{DEVEL_CODE_VAL_RE}\s*)*:end\1"
DEVEL_CODE_BLOCK_GENERIC_RE = r"([A-Za-z0-9_]+):(?:.*):end\1"

PARAM_VALUE_RE = re.compile(
    rf"""
^\s*(?P<key>[a-z_]+)
\s*[ \t:=]\s*
(?P<val>(?:\s*{NUMBER_RE.pattern})+|\S+)
(?P<unit>\s\S*\w\S*)?
\s*$
""",
    re.IGNORECASE | re.VERBOSE,
)

IONIC_CONSTRAINTS_RE = re.compile(rf"^\s*\d\s+{ATOM_RE}{THREEVEC_RE}")
POSITIONS_LINE_RE = re.compile(
    rf"^\s*(?P<spec>{SPECIES_RE})"
    rf"(?P<pos>(?:\s+{FLOAT_RAT_RE.pattern}){{3}})",
)
POSITIONS_SPIN_RE = re.compile(
    rf"(?:spin|magmom)\s*[= :\t]\s*(?P<spin>{FLOAT_RAT_RE.pattern})",
    re.IGNORECASE,
)
POSITIONS_MIXTURE_RE = re.compile(
    rf"mixture[= :\t]\(\s*{labelled_floats(('mix', 'ratio'))}\)",
    re.IGNORECASE,
)
SPEC_PROP_RE = re.compile(rf"\s*{ATOM_NAME_RE}\s+(?P<val>.*)")

# Berry Phase Polarisation
POL_HEADER_RE = re.compile(r"^\s*(?P<key>[\w\s]+)\s*\((?P<unit>[^)]+)\)")
POL_TABLE_RE = re.compile(
    r"\s*(?P<key>Ionic|Elec|Total)[^:]+:"
    rf"{labelled_floats(('a', 'b', 'c'))}\s*"
    r"\((?P<unit>[^)]+)\)",
)

# Elastic Constants file
ELASTIC_BLOCK_RE = re.compile(r"\s*BEGIN (?P<key>[^(]+)\((?P<unit>[^)]+)\)")
ELASTIC_INTERNAL_STRAIN_RE = re.compile(
    rf"\s*(?P<index>\d+)\s+(?P<spec>{ATOM_NAME_RE})\s+"
    r"(?P<dir>[XYZ])\s+"
    f"{labelled_floats(SND_D)}",
)



"""Module containing dumpers for formats."""
from __future__ import annotations

import json
import pprint
from contextlib import suppress
from typing import Any, Callable, TextIO

_YAML_TYPE = None

with suppress(ImportError):
    import yaml
    _YAML_TYPE = "pyyaml"

with suppress(ImportError):
    from ruamel import yaml as ruamel
    _YAML_TYPE = "ruamel"


#: Dumping function protocol.
Dumper = Callable[[Any, TextIO], None]


def json_dumper(data: Any, file: TextIO) -> None:
    """
    JSON format dumper.

    Parameters
    ----------
    data
        Data to dump.
    file
        File to dump to.
    """
    json.dump(data, file, indent=2)


def ruamel_dumper(data: Any, file: TextIO) -> None:
    """
    YAML (ruamel.yaml) format dumper.

    Parameters
    ----------
    data
        Data to dump.
    file
        File to dump to.
    """
    yaml_eng = ruamel.YAML(typ="unsafe")
    yaml_eng.dump(data, file)


def pyyaml_dumper(data: Any, file: TextIO) -> None:
    """
    YAML (pyyaml) format dumper.

    Parameters
    ----------
    data
        Data to dump.
    file
        File to dump to.
    """
    yaml.dump(data, file)


def pprint_dumper(data: Any, file: TextIO) -> None:
    """
    Pretty print dumper.

    Parameters
    ----------
    data
        Data to dump.
    file
        File to dump to.
    """
    print(pprint.pformat(data), file=file)


def print_dumper(data: Any, file: TextIO) -> None:
    """
    Python print dumper.

    Parameters
    ----------
    data
        Data to dump.
    file
        File to dump to.
    """
    print(data, file=file)


def get_dumpers(dump_fmt: str) -> Dumper:
    """
    Get appropriate dump for unified interface.

    Parameters
    ----------
    dump_fmt
        Formats to dump to.

    Returns
    -------
    Dumper
        Dumping function.

    Raises
    ------
    ValueError
        Invalid `dump_fmt` provided.
    ImportError
        No valid YAML dumper and `yaml` requested.

    See Also
    --------
    SUPPORTED_FORMATS
        Acceptable values for `dump_fmt`.
    """
    if dump_fmt == "yaml":
        if _YAML_TYPE is None:
            raise ImportError("Couldn't find valid yaml dumper (ruamel.yaml / yaml)"
                              "please install and try again.")
        dump_fmt = _YAML_TYPE

    if dump_fmt not in SUPPORTED_FORMATS:
        raise ValueError(f"Cannot output in {dump_fmt} format. "
                         f"Valid keys are: {', '.join(SUPPORTED_FORMATS.keys())}")

    return SUPPORTED_FORMATS[dump_fmt]


#: Currently supported dumpers.
SUPPORTED_FORMATS: dict[str, Dumper] = {"json": json_dumper,
                                        "ruamel": ruamel_dumper,
                                        "pyyaml": pyyaml_dumper,
                                        "pprint": pprint_dumper,
                                        "print": print_dumper}



# pylint: skip-file

import io
import pprint
import unittest
import pytest

from castep_outputs.parsers import parse_cell_param_file
from castep_outputs.parsers.cell_param_file_parser import _parse_pspot_string


class test_cell_parser(unittest.TestCase):

    def test_parse_lattice_abc(self):
        test_text = io.StringIO("""
%BLOCK LATTICE_ABC
    8.9780000000    5.7400000000	9.9690000000
    90.000000000    90.000000000	90.000000000
%ENDBLOCK LATTICE_ABC
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"lattice_abc": {"data": [(8.978, 5.74, 9.969),
                                                              (90.0, 90.0, 90.0)]}})

    def test_parse_lattice_cart(self):
        test_text = io.StringIO("""
%BLOCK LATTICE_CART
ang    # angstrom units
   6.314500000000000   0.000000000000000   0.000000000000000
   0.000000000000000   6.314500000000000   0.000000000000000
   0.000000000000000   0.000000000000000   6.314500000000000
%ENDBLOCK LATTICE_CART
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"lattice_cart": {"data": [(6.3145, 0.0, 0.0),
                                                               (0.0, 6.3145, 0.0),
                                                               (0.0, 0.0, 6.3145)],
                                                      "units": "ang"}})

    def test_parse_mixture(self):
        test_text = io.StringIO("""
%BLOCK POSITIONS_FRAC
    Al    0.2500000000    0.5000000000    0.0000000000 MIXTURE:(   1  0.666667)
    Al   -0.2500000000   -0.5000000000    0.0000000000 MIXTURE:(   2  0.666667)
    Al    0.2500000000    0.0000000000    0.5000000000 MIXTURE:(   3  0.666667)
    Si    0.2500000000    0.5000000000    0.0000000000 MIXTURE:(   1  0.333333)
    Si   -0.2500000000   -0.5000000000    0.0000000000 MIXTURE:(   2  0.333333)
    Si    1/4    0    1/2 MIXTURE:(   3  0.333333)
%ENDBLOCK POSITIONS_FRAC
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"positions_frac": {
            ("Al", 1): {"mix_index": 1,
                        "pos": (0.25, 0.5, 0.0),
                        "weight": 0.666667},
            ("Al", 2): {"mix_index": 2,
                        "pos": (-0.25, -0.5, 0.0),
                        "weight": 0.666667},
            ("Al", 3): {"mix_index": 3,
                        "pos": (0.25, 0.0, 0.5),
                        "weight": 0.666667},
            ("Si", 1): {"mix_index": 1,
                        "pos": (0.25, 0.5, 0.0),
                        "weight": 0.333333},
            ("Si", 2): {"mix_index": 2,
                        "pos": (-0.25, -0.5, 0.0),
                        "weight": 0.333333},
            ("Si", 3): {"mix_index": 3,
                        "pos": (0.25, 0.0, 0.5),
                        "weight": 0.333333}},
                                     },
                         )

    def test_parse_spin_magmom(self):
        test_text = io.StringIO("""
%BLOCK POSITIONS_FRAC
    Fe 0.00    0.00    0.00
    Fe 0.50    0.50    0.00 SPIN= 1/2
    Fe 0.50    0.00    0.50 SPIN : -2
    Fe 0.00    0.50    0.50 MAGMOM 4
%ENDBLOCK POSITIONS_FRAC
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"positions_frac": {("Fe", 1): {"pos": (0.0, 0.0, 0.0)},
                                                        ("Fe", 2): {"pos": (0.5, 0.5, 0.0),
                                                                    "spin": 0.5},
                                                        ("Fe", 3): {"pos": (0.5, 0.0, 0.5),
                                                                    "spin": -2.0},
                                                        ("Fe", 4): {"pos": (0.0, 0.5, 0.5),
                                                                    "spin": 4.0}}})

    def test_parse_position_abs(self):
        test_text = io.StringIO("""
%BLOCK POSITIONS_ABS
    O     6.2450000000   -2.3470000000    3.0440000000
    Al    5.7110000000    2.3470000000    5.0230000000
    Al    1    1/2  1/3
%ENDBLOCK POSITIONS_ABS
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"positions_abs": {("Al", 1): {"pos": (5.711, 2.347, 5.023),
                                                                   "units": "Ang"},
                                                       ("Al", 2): {"pos": (1.0, 1/2, 1/3),
                                                                   "units": "Ang"},
                                                       ("O", 1): {"pos": (6.245, -2.347, 3.044),
                                                                  "units": "Ang"}}})

    def test_parse_position_abs_units(self):
        test_text = io.StringIO("""
%BLOCK POSITIONS_ABS
Furlongs
    O     6.2450000000   -2.3470000000    3.0440000000
    Al    5.7110000000    2.3470000000    5.0230000000
    Al    1    1/2  1/3
%ENDBLOCK POSITIONS_ABS
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"positions_abs": {("Al", 1): {"pos": (5.711, 2.347, 5.023),
                                                                   "units": "Furlongs"},
                                                       ("Al", 2): {"pos": (1.0, 1/2, 1/3),
                                                                   "units": "Furlongs"},
                                                       ("O", 1): {"pos": (6.245, -2.347, 3.044),
                                                                  "units": "Furlongs"}}})

    def test_parse_cell_constraints(self):
        test_text = io.StringIO("""
%BLOCK CELL_CONSTRAINTS
       1       2       3
       0       0       0
%ENDBLOCK CELL_CONSTRAINTS
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"cell_constraints": {"data": [(1.0, 2.0, 3.0),
                                                                   (0.0, 0.0, 0.0)]}})

    def test_parse_ionic_constraints(self):
        test_text = io.StringIO("""
%BLOCK IONIC_CONSTRAINTS
       1       W       1    1.0000000000    0.0000000000    0.0000000000
       2       W       1    0.0000000000    1.0000000000    0.0000000000
       3       W       1    0.0000000000    0.0000000000    1.0000000000
       4       W       2    1.0000000000    0.0000000000    0.0000000000
%ENDBLOCK IONIC_CONSTRAINTS
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"ionic_constraints": {("W", 1): [(1.0, 0.0, 0.0),
                                                                      (0.0, 1.0, 0.0),
                                                                      (0.0, 0.0, 1.0)],
                                                           ("W", 2): [(1.0, 0.0, 0.0)]}})

    def test_parse_nonlinear_constraints(self):
        test_text = io.StringIO("""
%BLOCK NONLINEAR_CONSTRAINTS
distance       H  4  0  0  0       O  2  0  1  0
    bend       H  5  0  0  0       C  1  1  0  1       H  2  0  0  0
 torsion       H  6  0  0  0       H  3  1  0  0       H  1  0  0  1       H  9  1  1  0
%ENDBLOCK NONLINEAR_CONSTRAINTS
    """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"nonlinear_constraints": [{"key": "distance",
                                                                "atoms": {("H", 4): (0, 0, 0),
                                                                          ("O", 2): (0, 1, 0)},
                                                                },
                                                               {"key": "bend",
                                                                "atoms": {("C", 1): (1, 0, 1),
                                                                          ("H", 2): (0, 0, 0),
                                                                          ("H", 5): (0, 0, 0)},
                                                                },
                                                               {"key": "torsion",
                                                                "atoms": {("H", 1): (0, 0, 1),
                                                                          ("H", 3): (1, 0, 0),
                                                                          ("H", 6): (0, 0, 0),
                                                                          ("H", 9): (1, 1, 0)},
                                                                }]})

    def test_parse_external_pressure(self):
        test_text = io.StringIO("""
    %BLOCK EXTERNAL_PRESSURE
GPa
    5.0000000000    0.0000000000    0.0000000000
                    5.0000000000    0.0000000000
                                    5.0000000000
%ENDBLOCK EXTERNAL_PRESSURE

        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"external_pressure": {"data": [(5.0, 0.0, 0.0),
                                                                    (5.0, 0.0),
                                                                    (5.0,)],
                                                           "units": "GPa"}})

    def test_parse_weighted_kpoints(self):
        test_text = io.StringIO("""
%BLOCK SUPERCELL_KPOINT_LIST
    0.0000000000    0.0000000000    0.0000000000    0.0800000000
    0.0000000000    0.0833333333    0.0000000000    0.0400000000
    0.0000000000    0.1666666667    0.0000000000    0.0400000000
%ENDBLOCK SUPERCELL_KPOINT_LIST
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"supercell_kpoint_list": {"data": [(0.0, 0.0, 0.0, 0.08),
                                                                        (0.0, 0.0833333333, 0.0, 0.04),
                                                                        (0.0, 0.1666666667, 0.0, 0.04)]}})

    def test_parse_supercell_matrix(self):
        test_text = io.StringIO("""
%BLOCK PHONON_SUPERCELL_MATRIX
 2 2 0
 0 2 2
 2 0 2
%ENDBLOCK PHONON_SUPERCELL_MATRIX
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"phonon_supercell_matrix": {"data": [(2.0, 2.0, 0.0),
                                                                          (0.0, 2.0, 2.0),
                                                                          (2.0, 0.0, 2.0)]}})

    def test_parse_symops(self):
        test_text = io.StringIO("""
%BLOCK SYMMETRY_OPS
    # SYM OP 1
   -1.0000000000    0.0000000000    0.0000000000
    0.0000000000   -1.0000000000    0.0000000000
    0.0000000000    0.0000000000    1.0000000000
    0.5000000000    0.0000000000    0.5000000000
    # SYM OP 2
    1.0000000000    0.0000000000    0.0000000000
    0.0000000000    1.0000000000    0.0000000000
    0.0000000000    0.0000000000    1.0000000000
    0.5000000000    0.0000000000    0.5000000000
%ENDBLOCK SYMMETRY_OPS
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"symmetry_ops": [{"r": [(-1.0, 0.0, 0.0),
                                                             (0.0, -1.0, 0.0),
                                                             (0.0, 0.0, 1.0)],
                                                       "t": (0.5, 0.0, 0.5)},
                                                      {"r": [(1.0, 0.0, 0.0),
                                                             (0.0, 1.0, 0.0),
                                                             (0.0, 0.0, 1.0)],
                                                       "t": (0.5, 0.0, 0.5)}]})


    def test_parse_quantisation_axis(self):
        test_text = io.StringIO("""
QUANTIZATION_AXIS : 1 1 -1
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"quantization_axis": (1, 1, -1)})

    def test_parse_external_efield(self):
        test_text = io.StringIO("""
%BLOCK EXTERNAL_EFIELD
HARTREE/BOHR/E
0.0 1.0 0.1
%ENDBLOCK EXTERNAL_EFIELD
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"external_efield": {"data": [(0.0, 1.0, 0.1)],
                                                         "units": "HARTREE/BOHR/E"}})

    def test_parse_ionic_velocities(self):
        test_text = io.StringIO("""
%BLOCK IONIC_VELOCITIES
ang/ps
 1.000  2.000  1.000
 2.000  3.000  3.000
-2.000 -3.000 -3.000
-1.000 -2.000 -1.000
%ENDBLOCK IONIC_VELOCITIES
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"ionic_velocities": {"data": [(1.0, 2.0, 1.0),
                                                                   (2.0, 3.0, 3.0),
                                                                   (-2.0, -3.0, -3.0),
                                                                   (-1.0, -2.0, -1.0)],
                                                          "units": "ang/ps"}})

    def test_parse_kpoint_path(self):
        test_text = io.StringIO("""
%BLOCK BS_KPOINT_PATH
    0.3333333333    0.3750000000    0.3333333333
    0.3333333333    0.3750000000    0.0000000000
    1/3             1/8             1/3
%ENDBLOCK BS_KPOINT_PATH
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"bs_kpoint_path": {"data": [(0.3333333333, 0.375, 0.3333333333),
                                                                 (0.3333333333, 0.375, 0.0),
                                                                 (0.3333333333333333, 0.125, 0.3333333333333333),
                                                                 ]}})

    def test_parse_kpoint_spacing(self):
        test_text = io.StringIO("""
KPOINT_MP_SPACING = 0.25 1/Ang
BS_KPOINT_PATH_SPACING : 0.125
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"kpoint_mp_spacing": (0.25, "1/Ang"),
                                     "bs_kpoint_path_spacing": 0.125})

    def test_parse_fix(self):
        test_text = io.StringIO("""
FIX_ALL_CELL : F
FIX_ALL_IONS : T
FIX_COM = false
FIX_VOL TRUE
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"fix_all_cell": False,
                                     "fix_all_ions": True,
                                     "fix_com": False,
                                     "fix_vol": True})

    def test_parse_lcao(self):
        test_text = io.StringIO("""
%BLOCK SPECIES_LCAO_STATES
       O         1
      Al         2
      Ti:tag     3
%ENDBLOCK SPECIES_LCAO_STATES
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"species_lcao_states": {"data": {"Al": 2,
                                                                      "O": 1,
                                                                      "Ti:tag": 3}}})

    def test_get_species_mass(self):
        test_text = io.StringIO("""
%BLOCK SPECIES_MASS
       O     15.9989995956
      Al     26.9820003510
      Ti:Tag 47.9000015259
      Cs    132.9049987793
%ENDBLOCK SPECIES_MASS
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"species_mass": {"data": {"Al": 26.982000351,
                                                               "Cs": 132.9049987793,
                                                               "O": 15.9989995956,
                                                               "Ti:Tag": 47.9000015259}}})

    def test_parse_sedc_params(self):
        test_text = io.StringIO("""
%BLOCK SEDC_CUSTOM_PARAMS
H C6:0.0 R0:1.6404 alpha:0.6668
Ti:Tag C6:1.0 R0:3.6404 alpha:-0.6668
%ENDBLOCK SEDC_CUSTOM_PARAMS
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"sedc_custom_params": {"H": {"C6": 0.0,
                                                                  "R0": 1.6404,
                                                                  "alpha": 0.6668},
                                                            "Ti:Tag": {"C6": 1.0,
                                                                       "R0": 3.6404,
                                                                       "alpha": -0.6668}}})

    def test_parse_hubbard_u(self):
        test_text = io.StringIO("""
%BLOCK HUBBARD_U
  eV
    Sm 1   f: 6.1
    Ni     d: 2.4
    U 2  d: 1.2 f: 2.1
%ENDBLOCK HUBBARD_U
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"hubbard_u": {"units": "eV",
                                                   "Ni": {"d": 2.4},
                                                   ("Sm", 1): {"f": 6.1},
                                                   ("U", 2): {"d": 1.2, "f": 2.1}}})


class test_param_parser(unittest.TestCase):
    def test_int_params(self):
        test_text = io.StringIO("""
BACKUP_INTERVAL : 3600
RUN_TIME=360
NLXC_PPD_SIZE_X  4
NLXC_PPD_SIZE_Y = 2
NLXC_PPD_SIZE_Z  -2
        """)
        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"backup_interval": 3600,
                                     "nlxc_ppd_size_x": 4,
                                     "nlxc_ppd_size_y": 2,
                                     "nlxc_ppd_size_z": -2.0,
                                     "run_time": 360})

    def test_bool_params(self):
        test_text = io.StringIO("""
CALCULATE_STRESS : TRUE
CALCULATE_DENSDIFF = TRUE
CALCULATE_ELF  F
CALCULATE_HIRSHFELD : False
        """)
        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"calculate_densdiff": True,
                                     "calculate_elf": False,
                                     "calculate_hirshfeld": False,
                                     "calculate_stress": True})

    def test_string_params(self):
        test_text = io.StringIO("""
charge_unit: e
CHECKPOINT : test.check
CONTINUATION : DEFAULT
PSPOT_BETA_PHI_TYPE = REAL
SMEARING_SCHEME  ColdSmearing
File_param aa.bbb
""")
        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"charge_unit": "e",
                                     "checkpoint": "test.check",
                                     "continuation": "DEFAULT",
                                     "file_param": "aa.bbb",
                                     "pspot_beta_phi_type": "REAL",
                                     "smearing_scheme": "ColdSmearing"})

    def test_vector_params(self):
        test_text = io.StringIO("""
kpoints_mp_grid : 10 10 10
        """)
        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"kpoints_mp_grid": (10, 10, 10)})

    def test_physical_params(self):
        test_text = io.StringIO("""
geom_energy_tol : 0.00005 eV
GEOM_MODULUS_EST = 125.4 GPa
MIX_METRIC_Q : 20.0
        """)

        test_dict = parse_cell_param_file(test_text)[0]

        self.assertEqual(test_dict, {"geom_energy_tol": (5e-05, "eV"),
                                     "geom_modulus_est": (125.4, "GPa"),
                                     "mix_metric_q": 20.0})

    def test_devel_code(self):
        test_text = io.StringIO("""
%block devel_code
bool_value=F
true_val=True
int_value=31
string_value=Hello
float_number=15.12
PP: test_par=1 :ENDPP
%endblock devel_code
        """)
        test_dict = parse_cell_param_file(test_text)[0]
        self.assertEqual(test_dict, {"devel_code": {"pp": {"test_par": 1},
                                                    "bool_value": False,
                                                    "true_val": True,
                                                    "float_number": 15.12,
                                                    "int_value": 31,
                                                    "string_value": "Hello"}})

    def test_xc_def(self):
        test_text = io.StringIO("""
%block xc_definition
oep 1.0
hf 2.1
lda 0.1
LIBXC_MGGA_XC_ZLP 2.
NLXC_SCREENING_LENGTH 1.
NLXC_SCREENING_FUNCTION THOMAS-FERMI
NLXC_DIVERGENCE_CORR ON
NLXC_PPD_INT ON
%endblock xc_definition
""")
        test_dict = parse_cell_param_file(test_text)[0]
        pprint.pprint(test_dict)
        self.assertEqual(test_dict, {"xc_definition": {
            "xc": {
                "oep": 1.0,
                "hf": 2.1,
                "lda": 0.1,
                "libxc_mgga_xc_zlp": 2.0,
            },
            "params": {
                "nlxc_divergence_corr": True,
                "nlxc_ppd_int": True,
                "nlxc_screening_length": 1.,
                "nlxc_screening_function": "THOMAS-FERMI",
            },
        }})


@pytest.mark.parametrize("string", [
    "1|0.8|11|15|18|10N(qc=8)[]",
    "1|1.2|23|29|33|20N:21L(qc=9)[]",
    "2|1.4|1.4|1.3|6|10|12|20:21(qc=6)",
    "1|0.8|0.8|0.6|2|6|8|10(qc=6)",
    "2|1.0|1.3|0.7|13|16|18|20:21(qc=7)",
    "2|1.8|3.675|5.512|7.35|30UU:31UU:32LGG[]",
    "2|1.8|3.675|5.512|7.35|30UU:31UU:32LGG",
    "2|1.9|2.1|1.3|3.675|5.512|7.35|30N1.9:31N2.1:32N2.1(tm,nonlcc)[]",
    "2|1.75|1.75|1.3|6.4|8.1|11|30N:31N:32L(qc=6.5)[3s2,3p1.5,3d0.25]",
    "2|1.8|1.8|1.3|2|3|4|30:31:32LGG(qc=4)",
    "2|1.8|3.675|5.512|7.35|30UU:31UU:32LGG[]",
    "2|1.8|3.675|5.512|7.35|30UU:31UU:32LGG{1s1}[]",
])
def test_pspot_parser(string):
    pytest.skip("Correct answers not provided")
    test_dict = _parse_pspot_string(string)

    pprint.pprint(test_dict)
    # self.assertEqual(test_dict, {})


if __name__ == "__main__":
    unittest.main()



# pylint: disable=too-many-lines, too-many-branches, too-many-statements
"""Functions generally used in parsing castep files."""
from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Sequence

from ..utilities import castep_res as REs
from ..utilities.castep_res import get_numbers
from ..utilities.filewrapper import Block
from ..utilities.utility import (
    atreg_to_index,
    fix_data_types,
    stack_dict,
    to_type,
)


def parse_regular_header(block: Block,
                         extra_opts: Sequence[str] = ()) -> dict[str, float | int]:
    """
    Parse (semi-)standard castep file header block (given as iterable over lines).

    Parameters
    ----------
    block
        Block to parse.
    extra_opts
        Extra regexes to match (stored as floats).

    Returns
    -------
    dict[str, float | int]
        Parsed header block.
    """
    data = {}
    coords = defaultdict(list)
    for line in block:
        if line.strip().startswith("Number of"):
            _, _, *key, val = line.split()
            data[" ".join(key)] = int(float(val))
        elif "Unit cell vectors" in line:
            data["unit_cell"] = [to_type(next(block).split(), float)
                                 for _ in range(3)]

        elif match := REs.ATOMIC_DATA_3VEC.search(line):
            ind = atreg_to_index(match)
            coords[ind] = to_type(match.group("x", "y", "z"), float)

        elif match := REs.FRACCOORDS_RE.match(line):
            stack_dict(coords, match.groupdict())

        elif match := re.search(f"({'|'.join(extra_opts)})", line):
            data[match.group(0)] = to_type(get_numbers(line), float)

    fix_data_types(coords, {"index": int,
                            "u": float,
                            "v": float,
                            "w": float,
                            "mass": float})
    data["coords"] = coords
    return data



"""Module to assist processing of parsed outputs."""

from typing import NamedTuple, TypeVar

from ..utilities.castep_res import get_atom_parts

Self = TypeVar("Self", bound="AtomLabel")


class AtomLabel(NamedTuple):
    """Standard castep atom label."""

    #: Atom species.
    species: str
    #: Index into positions block.
    index: int
    #: Extra tag information.
    tag: str = ""
    #: Overridden internal name via label.
    label: str = ""

    @classmethod
    def from_str(cls, string: str) -> Self:
        """
        Build a label from a key-string.

        Parameters
        ----------
        string
            String to parse.

        Returns
        -------
        AtomLabel
            Processed string.
        """
        return cls(**get_atom_parts(string))



"""Parse castep .md or .geom files."""
from __future__ import annotations

from collections import defaultdict
from typing import TextIO, TypedDict

from ..utilities.castep_res import ATOMIC_DATA_TAG, TAG_RE, get_numbers
from ..utilities.constants import FST_D, TAG_ALIASES
from ..utilities.datatypes import AtomIndex, ThreeByThreeMatrix, ThreeVector
from ..utilities.filewrapper import Block
from ..utilities.utility import add_aliases, atreg_to_index, file_or_path, to_type


class MDAtomProps(TypedDict):
    """Atom properties on MD and GeomOpt."""

    #: Force on atom in Ha/Bohr.
    force: ThreeVector
    #: Position of atom in Bohr.
    position: ThreeVector
    #: Velocity of atom in Bohr/aut.
    velocity: ThreeVector
    #: Alias of :attr:`force`
    F: ThreeVector
    #: Alias of :attr:`position`
    R: ThreeVector
    #: Alias of :attr:`velocity`
    V: ThreeVector


class MDGeomTimestepInfo(TypedDict, total=False):
    """
    MD and GeomOpt output info.

    Notes
    -----
    Also contains :any:`AtomIndex` keys to per-atom information.
    """

    #: Elapsed MD Time.
    time: float
    #: Current energies: total, potential, kinetic.
    energy: tuple[float, float, float]
    #: Instantaneous temperature.
    temperature: tuple[float]
    #: Hydrostatic pressure.
    pressure: tuple[float]
    #: Current cell vectors.
    lattice_vectors: ThreeByThreeMatrix
    #: Current cell changes.
    lattice_velocity: ThreeByThreeMatrix
    #: Current stresses.
    stress: ThreeByThreeMatrix
    #: Atomic properties
    ions: dict[AtomIndex, MDAtomProps]

    #: Alias of :attr:`energy`.
    E: list[float]
    #: Alias of :attr:`temperature`.
    T: list[float]
    #: Alias of :attr:`pressure`.
    P: list[float]
    #: Alias of :attr:`lattice_vectors`.
    h: ThreeByThreeMatrix
    #: Alias of :attr:`lattice_velocity`.
    hv: ThreeByThreeMatrix
    #: Alias of :attr:`stress`.
    S: ThreeByThreeMatrix


def parse_md_geom_frame(block: Block) -> MDGeomTimestepInfo:
    """
    Parse a single frame of a .md/.geom file.

    Parameters
    ----------
    block
        Block containing frame of data.

    Returns
    -------
    MDGeomTimestepInfo
        Parsed frame of data.
    """
    curr: MDGeomTimestepInfo = defaultdict(list)
    curr["ions"] = {}

    for line in block:
        if not line.strip():
            pass
        elif not TAG_RE.search(line):  # Timestep
            curr["time"] = to_type(get_numbers(line)[0], float)

        elif match := ATOMIC_DATA_TAG.match(line):
            ion = atreg_to_index(match)
            curr["ions"].setdefault(ion, {})
            curr["ions"][ion][match.group("tag")] = to_type([match.group(d) for d in FST_D], float)

        elif match := TAG_RE.search(line):
            curr[match.group("tag")].append([*to_type(get_numbers(line), float)])

    add_aliases(curr, TAG_ALIASES)
    for ion in curr["ions"].values():
        add_aliases(ion, TAG_ALIASES)

    return curr


@file_or_path(mode="r")
def parse_md_geom_file(md_geom_file: TextIO) -> list[MDGeomTimestepInfo]:
    """
    Parse standard .md and .geom files.

    Parameters
    ----------
    md_geom_file
        Open handle to file to parse.

    Returns
    -------
    list[MDGeomTimestepInfo]
        Step-by-step Parsed info.
    """
    while "END header" not in md_geom_file.readline():
        pass
    md_geom_file.readline()
    steps = []
    while block := Block.from_re("", md_geom_file, "", "^$", eof_possible=True):
        steps.append(parse_md_geom_frame(block))

    return steps


# Aliases
parse_md_file = parse_md_geom_file
parse_geom_file = parse_md_geom_file



"""Parse castep .bands files."""
from __future__ import annotations

import re
from typing import Literal, TextIO, TypedDict

from ..utilities import castep_res as REs
from ..utilities.datatypes import ThreeVector
from ..utilities.filewrapper import Block
from ..utilities.utility import file_or_path, to_type
from .parse_utilities import parse_regular_header


class BandsQData(TypedDict, total=False):
    """Per k-point info of band."""

    #: List of band eigenvalues.
    band: ThreeVector
    #: List of eigenvalues for up component of band.
    band_up: ThreeVector
    #: List of eigenvalues for down component of band.
    band_down: ThreeVector
    #: Position in space.
    qpt: ThreeVector
    #: Current spin component.
    spin_comp: int
    #: K point weight.
    weight: float


BandsFileInfo = TypedDict("BandsFileInfo", {
    "eigenvalues": int,
    "electrons": int,
    "k-points": int,
    "spin components": int,
    "Fermi Energy": float,
    "coords": dict,
    "bands": list[BandsQData],
    })


@file_or_path(mode="r")
def parse_bands_file(bands_file: TextIO) -> BandsFileInfo:
    """
    Parse castep .bands file.

    Parameters
    ----------
    bands_file
        Open handle to file to parse.

    Returns
    -------
    BandsFileInfo
        Parsed info.
    """
    bands_info: BandsFileInfo = {"bands": []}
    qdata: BandsQData = {}
    accum: list[str] = []
    current: Literal["band", "band_up", "band_down"] = "band"

    block = Block.from_re("", bands_file, "", REs.THREEVEC_RE, n_end=3)
    data = parse_regular_header(block, ("Fermi energy",))
    bands_info.update(data)

    for line in bands_file:
        if line.startswith("K-point"):
            if qdata:
                qdata[current] = to_type(accum, float)
                qdata["spin_comp"] = "band_down" in qdata
                bands_info["bands"].append(qdata)
                accum = []
                current = "band"
            _, _, *qpt, weight = line.split()
            qdata = {"qpt": to_type(qpt, float), "weight": float(weight)}

        elif line.startswith("Spin component"):
            spin_comp = int(line.split()[2])
            if spin_comp != 1:
                qdata["band_up"] = to_type(accum, float)
                accum = []
                current = "band_down"

        elif re.match(rf"^\s*{REs.FNUMBER_RE}$", line.strip()):
            accum.append(line.strip())

    if qdata:
        qdata[current] = to_type(accum, float)
        qdata["spin_comp"] = "band_down" in qdata
        bands_info["bands"].append(qdata)

    return bands_info



"""Utility functions for parsing castep outputs."""

from __future__ import annotations

import collections.abc
import fileinput
import functools
import logging
import re
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator, MutableMapping, Sequence
from copy import copy
from fractions import Fraction
from functools import partial, singledispatch, wraps
from itertools import filterfalse
from pathlib import Path
from struct import unpack
from typing import Any, Literal, TextIO, TypeVar

import castep_outputs.utilities.castep_res as REs

from ..utilities.filewrapper import Block
from .filewrapper import FileWrapper

T = TypeVar("T")
NORMALISE_RE = re.compile(r"[_\W]+")
LoggingLevels = Literal["debug", "info", "warning", "error", "critical"]
Logger = Callable[[str, tuple[Any, ...], LoggingLevels], None]


def normalise_string(string: str) -> str:
    """
    Normalise a string.

    This includes:

    - Removing leading/trailing whitespace.
    - Making all spacing single-space.

    Parameters
    ----------
    string
        String to process.

    Returns
    -------
    str
        Normalised string.

    Examples
    --------
    >>> normalise_string(" Several   words  ")
    'Several words'
    """
    return " ".join(string.strip().split())


def normalise_key(string: str) -> str:
    """
    Normalise a dictionary key.

    This includes:

    - Removing all punctuation.
    - Lower-casing all.
    - Making all spacing single-underscore.

    Parameters
    ----------
    string
        String to process.

    Returns
    -------
    str
        Normalised string.

    Examples
    --------
    >>> normalise_key(" Several   words  ")
    'several_words'
    >>> normalise_key("A sentence.")
    'a_sentence'
    >>> normalise_key("I<3;;semi-colons;;!!!")
    'i_3_semi_colons'
    """
    return NORMALISE_RE.sub("_", string).strip("_").lower()


def atreg_to_index(dict_in: dict[str, str] | re.Match, *, clear: bool = True) -> tuple[str, int]:
    """
    Transform a matched atreg value to species index tuple.

    Optionally clear value from dictionary for easier processing.

    Parameters
    ----------
    dict_in
        Atreg to process.
    clear
        Whether to remove from incoming dictionary.

    Returns
    -------
    species : str
        Atomic species.
    ind : int
        Internal index.

    Examples
    --------
    >>> parsed_line = {'x': 3.1, 'y': 2.1, 'z': 1.0, 'spec': 'Ar', 'index': '1'}
    >>> atreg_to_index(parsed_line, clear=False)
    ('Ar', 1)
    >>> parsed_line
    {'x': 3.1, 'y': 2.1, 'z': 1.0, 'spec': 'Ar', 'index': '1'}
    >>> atreg_to_index(parsed_line)
    ('Ar', 1)
    >>> parsed_line
    {'x': 3.1, 'y': 2.1, 'z': 1.0}
    """
    spec, ind = dict_in["spec"], dict_in["index"]

    if isinstance(dict_in, dict) and clear:
        del dict_in["spec"]
        del dict_in["index"]

    return (spec, int(ind))


def normalise(obj: T, mapping: dict[type, type | Callable]) -> T:
    """
    Standardise data after processing.

    Recursively converts:

    - ``list`` s to ``tuple`` s
    - ``defaultdict`` s to ``dict`` s
    - types in `mapping` to their mapped type or apply mapped function.

    Parameters
    ----------
    obj
        Object to normalise.
    mapping
        Mapping of `type` to a callable transformation
        including class constructors.

    Returns
    -------
    Any
        Normmalised data.
    """
    if isinstance(obj, (tuple, list)):
        obj = tuple(normalise(v, mapping) for v in obj)
    elif isinstance(obj, (dict, defaultdict)):
        obj = {key: normalise(val, mapping) for key, val in obj.items()}

    if type(obj) in mapping:
        obj = mapping[type(obj)](obj)

    return obj


def json_safe(obj: dict | complex | T) -> dict | T:
    """
    Recursively transform datatypes into JSON safe variants.

    Including:

    - Ensuring dict keys are strings without spaces.
    - Ensuring complex numbers are split into real/imag components.

    Parameters
    ----------
    obj
        Incoming datatype.

    Returns
    -------
    Any
        Safe datatype.

    Examples
    --------
    >>> json_safe(3 + 4j)
    {'real': 3.0, 'imag': 4.0}
    >>> json_safe({('Ar', 'Sr'): 3})
    {'Ar_Sr': 3}
    >>> json_safe({(('Ar', 1), ('Sr', 1)): 3})
    {'Ar_1_Sr_1': 3}
    """
    if isinstance(obj, dict):
        obj_out = {}

        for key, val in obj.items():
            if isinstance(key, (tuple, list)):
                # Key in bonds is tuple[tuple[AtomIndex]]
                if isinstance(key[0], tuple):
                    key = "_".join(str(y) for x in key for y in x)
                else:
                    key = "_".join(map(str, key))
            obj_out[key] = json_safe(val)

        obj = obj_out
    elif isinstance(obj, complex):
        obj = {"real": obj.real, "imag": obj.imag}
    return obj


def flatten_dict(dictionary: MutableMapping[Any, Any],
                 parent_key: str = "",
                 separator: str = "_") -> dict[str, Any]:
    """
    Turn a nested dictionary into a flattened dictionary.

    Parameters
    ----------
    dictionary
        The dictionary to flatten.
    parent_key
        The string to prepend to dictionary's keys.
    separator
        The string used to separate flattened keys.

    Returns
    -------
    dict[str, Any]
        A flattened dictionary.

    Notes
    -----
    Taken from:
    https://stackoverflow.com/a/62186053

    Examples
    --------
    >>> flatten_dict({'hello': ['is', 'me'],
    ...               "goodbye": {"nest": "birds", "child": "moon"}})
    {'hello_0': 'is', 'hello_1': 'me', 'goodbye_nest': 'birds', 'goodbye_child': 'moon'}
    """
    items: list[tuple[Any, Any]] = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + separator + key if parent_key else key
        if isinstance(value, collections.abc.MutableMapping):
            items.extend(flatten_dict(value, new_key, separator).items())
        elif isinstance(value, (list, tuple)):
            for keyx, val in enumerate(value):
                items.extend(flatten_dict({str(keyx): val}, new_key).items())
        else:
            items.append((new_key, value))
    return dict(items)


def stack_dict(out_dict: dict[Any, list], in_dict: dict[Any, list]) -> None:
    """
    Append items in `in_dict` to the keys in `out_dict`.

    Parameters
    ----------
    out_dict
        Dict to append to.
    in_dict
        Dict to append from.
    """
    for key, val in in_dict.items():
        out_dict[key].append(val)


def add_aliases(in_dict: dict[str, Any],
                alias_dict: dict[str, str], *,
                replace: bool = False,
                inplace: bool = True) -> dict[str, Any]:
    """
    Add aliases of known names into dictionary.

    If replace is `True`, this will remove the original.

    Parameters
    ----------
    in_dict
        Dictionary of data to alias.
    alias_dict
        Mapping of from->to for keys in `in_dict`.
    replace
        Whether to remove the `from` key from `in_dict`.
    inplace
        Whether to return a copy or overwrite `in_dict`.

    Returns
    -------
    dict[str, Any]
        `in_dict` with keys substituted.

    Examples
    --------
    >>> add_aliases({'hi': 1, 'bye': 2}, {'hi': 'frog'})
    {'hi': 1, 'bye': 2, 'frog': 1}
    >>> add_aliases({'hi': 1, 'bye': 2}, {'hi': 'frog'}, replace=True)
    {'bye': 2, 'frog': 1}
    """
    out_dict = in_dict if inplace else copy(in_dict)

    for frm, new in alias_dict.items():
        if frm in in_dict:
            out_dict[new] = in_dict[frm]
            if replace:
                out_dict.pop(frm)
    return out_dict


@functools.singledispatch
def log_factory(file: TextIO | fileinput.FileInput | FileWrapper) -> Logger:
    """
    Return logging function to add file info to logs.

    Parameters
    ----------
    file : ~typing.TextIO or ~fileinput.FileInput or FileWrapper
        File to apply logging for.

    Returns
    -------
    Callable
        Function for logging data.
    """
    if hasattr(file, "name"):
        def log_file(message: str, *args: Any, level: LoggingLevels = "info") -> None:
            getattr(logging, level)(f"[{file.name}] {message}", *args)
    else:
        def log_file(message: str, *args: Any, level: LoggingLevels = "info") -> None:
            getattr(logging, level)(message, *args)

    return log_file


@log_factory.register
def _(file: fileinput.FileInput) -> Logger:
    def log_file(message: str, *args: Any, level: LoggingLevels = "info") -> None:
        getattr(logging, level)(f"[{file.filename()}:{file.lineno()}]"
                                f" {message}", *args)

    return log_file


@log_factory.register
def _(file: FileWrapper) -> Logger:
    def log_file(message: str, *args: Any, level: LoggingLevels = "info") -> None:
        getattr(logging, level)(f"[{file.name}:{file.lineno}]"
                                f" {message}", *args)

    return log_file


def determine_type(data: str) -> type:
    """
    Determine the datatype and return the appropriate type.

    For dealing with miscellaneous data read from input files.

    Parameters
    ----------
    data
        String to process.

    Returns
    -------
    type
        Best type to attempt.

    Examples
    --------
    >>> determine_type('T')
    <class 'bool'>
    >>> determine_type('False')
    <class 'bool'>
    >>> determine_type('3.1415')
    <class 'float'>
    >>> determine_type('123')
    <class 'int'>
    >>> determine_type('1/3')
    <class 'float'>
    >>> determine_type('BEEF')
    <class 'str'>
    """
    if data.title() in {"T", "True", "F", "False"}:
        return bool

    if re.match(rf"(?:\s*{REs.EXPFNUMBER_RE})+", data):
        return float

    if re.match(rf"(?:\s*{REs.RATIONAL_RE})+", data):
        return float

    if re.match(rf"(?:\s*{REs.INTNUMBER_RE})+", data):
        return int

    return str


def parse_int_or_float(numbers: Iterable[str]) -> int | float:
    """
    Parse numbers to `int` if all elements ints or `float` otherwise.

    Parameters
    ----------
    numbers
        Sequence of numbers to parse.

    Returns
    -------
    int or float
        Parsed numerical value.

    Examples
    --------
    >>> parse_int_or_float("3.141")
    3.141
    >>> parse_int_or_float("7")
    7
    """
    if all(x.isdigit() for x in numbers):
        return to_type(numbers, int)

    return to_type(numbers, float)


def _parse_float_or_rational(val: str) -> float:
    """
    Parse a float or rational to float.

    Parameters
    ----------
    val
        Value to parse.

    Returns
    -------
    float
        Parsed value.

    Examples
    --------
    >>> _parse_float_or_rational("0.5")
    0.5
    >>> _parse_float_or_rational("1/2")
    0.5
    """
    return float(Fraction(val))


def _parse_logical(val: str) -> bool:
    """
    Parse a logical to a bool.

    Parameters
    ----------
    val
        String to parse.

    Returns
    -------
    bool
        Parsed value.

    Notes
    -----
    Case-insensitive.

    Examples
    --------
    >>> _parse_logical("T")
    True
    >>> _parse_logical("TrUe")
    True
    >>> _parse_logical("1")
    True
    >>> _parse_logical("F")
    False
    """
    return val.title() in {"T", "True", "1"}


def _parse_float_bytes(val: bytes) -> float | Sequence[float]:
    r"""Parse (big-endian) bytes to float.

    Parameters
    ----------
    val : bytes
        Values to parse.

    Returns
    -------
    float | Sequence[float]
        Parsed value or list of values.

    Examples
    --------
    >>> one = b"?\xf0\x00\x00\x00\x00\x00\x00"
    >>> zero = b"\x00\x00\x00\x00\x00\x00\x00\x00"
    >>> _parse_float_bytes(zero)
    0.0
    >>> _parse_float_bytes(one)
    1.0
    >>> _parse_float_bytes(one*3)
    (1.0, 1.0, 1.0)
    """
    result = unpack(f">{len(val) // 8}d", val)
    return result if len(result) != 1 else result[0]


def _parse_int_bytes(val: bytes) -> int | Sequence[int]:
    r"""Parse (big-endian) bytes to int.

    Parameters
    ----------
    val : bytes
        Values to parse.

    Returns
    -------
    int | Sequence[int]
        Parsed value or list of values.

    Examples
    --------
    >>> one = b"\x00\x00\x00\x01"
    >>> zero = b"\x00\x00\x00\x00"
    >>> _parse_int_bytes(zero)
    0
    >>> _parse_int_bytes(one)
    1
    >>> _parse_int_bytes(one*3)
    (1, 1, 1)
    """
    result = unpack(f">{len(val) // 4}i", val)
    return result if len(result) != 1 else result[0]


def _parse_bool_bytes(val: bytes) -> bool | Sequence[bool]:
    r"""Parse (big-endian) bytes to bool.

    Parameters
    ----------
    val : bytes
        Values to parse.

    Returns
    -------
    bool | Sequence[bool]
        Parsed value or list of values.

    Examples
    --------
    >>> one = b"\x00\x00\x00\x01"
    >>> zero = b"\x00\x00\x00\x00"
    >>> _parse_bool_bytes(zero)
    False
    >>> _parse_bool_bytes(one)
    True
    >>> _parse_bool_bytes(one*3)
    (True, True, True)
    """
    result = tuple(map(bool, unpack(f">{len(val) // 4}i", val)))
    return result if len(result) != 1 else result[0]


def _parse_complex_bytes(val: bytes) -> complex | Sequence[complex]:
    r"""Parse (big-endian) bytes to complex.

    Parameters
    ----------
    val : bytes
        Values to parse.

    Returns
    -------
    complex | Sequence[complex]
        Parsed value or list of values.

    Examples
    --------
    >>> one = b"?\xf0\x00\x00\x00\x00\x00\x00"
    >>> zero = b"\x00\x00\x00\x00\x00\x00\x00\x00"
    >>> _parse_complex_bytes(zero + one)
    1j
    >>> _parse_complex_bytes(one + zero)
    (1+0j)
    >>> _parse_complex_bytes((one+one)*3)
    ((1+1j), (1+1j), (1+1j))
    """
    tmp = unpack(f">{len(val) // 8}d", val)
    result = tuple(map(complex, tmp[::2], tmp[1::2]))
    return result if len(result) != 1 else result[0]


_TYPE_PARSERS: dict[type, Callable] = {float: _parse_float_or_rational,
                                       bool: _parse_logical}
_BYTE_PARSERS: dict[type, Callable] = {complex: _parse_complex_bytes,
                                       float: _parse_float_bytes,
                                       bool: _parse_bool_bytes,
                                       int: _parse_int_bytes,
                                       str: partial(str, encoding="ascii")}


@functools.singledispatch
def to_type(data_in: T, _typ: type) -> T:
    """
    Convert types to `typ` regardless of if data_in is iterable or otherwise.

    Parameters
    ----------
    data_in : str or ~collections.abc.Sequence
        Data to convert.
    typ : type
        Type to convert to.

    Returns
    -------
    `typ` or tuple[`typ`, ...]
        Converted data.
    """
    return data_in


@to_type.register(str)
def _(data_in: str, typ: type[T]) -> T:
    parser: Callable[[str], T] = _TYPE_PARSERS.get(typ, typ)
    return parser(data_in)


@to_type.register(tuple)
@to_type.register(list)
def _(data_in: Any, typ: type[T]) -> tuple[T, ...]:
    parse_dict = _BYTE_PARSERS if data_in and isinstance(data_in[0], bytes) else _TYPE_PARSERS
    parser: Callable[[str], T] = parse_dict.get(typ, typ)
    return tuple(parser(x) for x in data_in)


@to_type.register(bytes)
def _(data_in: Any, typ: type[T]) -> T | tuple[T, ...]:
    parser: Callable[[bytes], T] = _BYTE_PARSERS.get(typ, typ)
    return parser(data_in)


def fix_data_types(in_dict: MutableMapping[str, Any], type_dict: dict[str, type]) -> None:
    """
    Apply correct types to elements of `in_dict` by mapping given in `type_dict`.

    Parameters
    ----------
    in_dict
        Dictionary of {key: values} to convert.
    type_dict
        Mapping of keys to types the keys should be converted to.

    See Also
    --------
    to_type : Conversion function.

    Notes
    -----
    Modifies the dictionary in-place.

    Examples
    --------
    >>> my_dict = {"int": "7", "float": "3.141", "bool": "T",
    ...            "vector": ["3", "4", "5"], "blank": "Hello"}
    >>> type_map = {"int": int, "float": float, "bool": bool, "vector": float}
    >>> fix_data_types(my_dict, type_map)
    >>> print(my_dict)
    {'int': 7, 'float': 3.141, 'bool': True, 'vector': (3.0, 4.0, 5.0), 'blank': 'Hello'}
    """
    for key, typ in type_dict.items():
        if key in in_dict:
            in_dict[key] = to_type(in_dict[key], typ)


def _strip_inline_comments(
        data: TextIO | FileWrapper | Block,
        *,
        comment_char: set[str],
) -> Iterator[str]:
    r"""
    Strip all comments from provided data.

    Parameters
    ----------
    data
        Data to strip comments from.
    comment_char
        Characters to interpret as comments.

    Yields
    ------
    str
        Data with line-initial comments stripped.

    Notes
    -----
    Also strips trailing, but not leading whitespace to clean up comment blocks.

    Also strips empty lines.

    Examples
    --------
    >>> from io import StringIO
    >>> inp = StringIO('''
    ... Hello
    ... # Initial line comment
    ... End of line # comment
    ... ''')
    >>> '|'.join(_strip_inline_comments(inp, comment_char={"#",}))
    'Hello|End of line'
    """
    comment_re = re.compile(f"({'|'.join(comment_char)})")

    for line in data:
        line = comment_re.split(line)[0].rstrip()
        if not line:
            continue

        yield line


def _strip_initial_comments(
        data: TextIO | FileWrapper | Block,
        *,
        comment_char: set[str],
) -> Iterator[str]:
    r"""
    Strip line-initial comments from provided data.

    Parameters
    ----------
    data
        Data to strip comments from.
    comment_char
        Characters to interpret as comments.

    Yields
    ------
    str
        Data with line-initial comments stripped.

    Notes
    -----
    Also strips trailing, but not leading whitespace to clean up comment blocks.

    Also strips empty lines.

    Examples
    --------
    >>> from io import StringIO
    >>> inp = StringIO('''
    ... Hello
    ... # Initial line comment
    ... End of line # comment
    ... ''')
    >>> '|'.join(_strip_initial_comments(inp, comment_char={"#",}))
    'Hello|End of line # comment'
    """
    comment_re = re.compile(rf"^\s*({'|'.join(comment_char)})")
    data = filterfalse(comment_re.match, data)
    data = map(str.rstrip, data)
    data = filter(None, data)
    yield from data


def strip_comments(
        data: TextIO | FileWrapper | Block,
        *,
        comment_char: str | set[str] = "#!",
        remove_inline: bool = False) -> Block:
    r"""
    Strip comments from data.

    Parameters
    ----------
    data
        Data to strip comments from.
    remove_inline
        Whether to remove inline comments or just line initial.
    comment_char
        Character sets to read as comments and remove.

        .. note::

            If the chars are passed as a string, it is assumed that
            each character is a comment character.

            To match a multicharacter comment you **must** pass this
            as a set or sequence of strings.

    Returns
    -------
    Block
        Block of data without comments.

    Notes
    -----
    Also strips trailing, but not leading whitespace to clean up comment blocks.

    Also strips empty lines.

    Examples
    --------
    >>> from io import StringIO
    >>> inp = StringIO('''
    ... Hello
    ... # Initial line comment
    ... End of line # comment
    ... // C-style
    ... ''')
    >>> x = strip_comments(inp, remove_inline=False)
    >>> type(x).__name__
    'Block'
    >>> '|'.join(x)
    'Hello|End of line # comment|// C-style'
    >>> _ = inp.seek(0)
    >>> x = strip_comments(inp, remove_inline=True)
    >>> '|'.join(x)
    'Hello|End of line|// C-style'
    >>> _ = inp.seek(0)
    >>> x = strip_comments(inp, comment_char={"//", "#"})
    >>> '|'.join(x)
    'Hello|End of line # comment'
    """
    if not isinstance(comment_char, set):
        comment_char = set(comment_char)

    strip_function = _strip_inline_comments if remove_inline else _strip_initial_comments
    stripped_comments = strip_function(data, comment_char=comment_char)

    return Block.from_iterable(stripped_comments, parent=data)


def get_only(seq: Sequence[T]) -> T:
    """
    Get the only element of a Sequence ensuring uniqueness.

    Parameters
    ----------
    seq
        Sequence of one element.

    Returns
    -------
    Any
        The sole element of the sequence.

    Raises
    ------
    ValueError
        Value is not alone.
    """
    val, *rest = seq
    if rest:
        raise ValueError(f"Multiple elements in sequence (remainder={', '.join(rest)}).")

    return val


def file_or_path(*, mode: Literal["r", "rb"], **open_kwargs: Any) -> Callable:  # paramspec 3.10+
    """Decorate to allow a parser to accept either a path or open file.

    Parameters
    ----------
    mode : Literal["r", "rb"]
        Open mode if passed a :class:`~pathlib.Path` or :class:`str`.

    Returns
    -------
    Callable
        Wrapped function able to handle open files or paths invisibly.
    """
    def inner(func: Callable) -> Callable:
        @wraps(func)
        def wrapped(file: str | Path, *args: Any, **kwargs: Any) -> Callable:
            file = Path(file)
            with file.open(mode, **open_kwargs) as in_file:
                return func(in_file, *args, **kwargs)

        func = singledispatch(func)
        func.register(str, wrapped)
        func.register(Path, wrapped)
        return func

    return inner



"""Test dumping works correcly."""

import io
import json
from functools import partial
from pathlib import Path

import pytest

from castep_outputs.cli.castep_outputs_main import parse_all, parse_single
from castep_outputs.utilities.utility import normalise

_DATA_FOLDER = Path(__file__).parent / "data_files"

def _to_complex(complx):
    if (isinstance(complx, dict) and not complx.keys() ^ {"real", "imag"}):
        return complex(**complx)
    return complx

@pytest.mark.parametrize("file,typ,ref", [
    ("test.cell", "cell", "cell"),
    ("test.param", "param", "param"),
    ("test.castep", "castep", "castep"),
    ("pp-md.castep", "castep", "pp-md"),
    ("si8-md.castep", "castep", "si8-md"),
    ("test.bands", "bands", "bands"),
    ("test.md", "md", "md"),
    ("test.elastic", "elastic", "elastic"),
    ("test.ts", "ts", "ts"),
    ("test.chdiff_fmt", "chdiff_fmt", "chdiff_fmt"),
    ("test.den_fmt", "den_fmt", "den_fmt"),
    ("test.pot_fmt", "pot_fmt", "pot_fmt"),
    ("test.elf_fmt", "elf_fmt", "elf_fmt"),
    ("test.xrd_sf", "xrd_sf", "xrd_sf"),
    ("test.phonon_dos", "phonon_dos", "phonon_dos"),
    ("test.efield", "efield", "efield"),
    ("test.magres", "magres", "magres"),
    ("test.tddft", "tddft", "tddft"),
    ("test.err", "err", "err"),
    ("test.phonon", "phonon", "phonon"),
    ("test.cst_esp", "cst_esp", "cst_esp"),
    ("test.epme", "epme", "epme"),
])
@pytest.mark.parametrize("out_format", ("json", "ruamel", "pyyaml"))
def test_dump(file, typ, ref, out_format):
    """Test dumpers work correctly."""
    # Get loaders
    if out_format == "json":
        loader = json.load
    elif out_format == "ruamel":
        module = pytest.importorskip("ruamel.yaml")
        yaml_eng = module.YAML(typ="unsafe")
        loader = yaml_eng.load
    elif out_format == "pyyaml":
        module = pytest.importorskip("yaml")
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader

        loader = partial(module.load, Loader=Loader)

    file = _DATA_FOLDER / file

    # Load reference data
    ref_file = _DATA_FOLDER / f"{ref}.{out_format}"
    with ref_file.open(encoding="utf-8") as test_file:
        ref_dict = loader(test_file)
    ref_dict = normalise(ref_dict, {dict: _to_complex})

    # Check produced data files identical
    test = io.StringIO()
    parse_all(test, out_format=out_format, **{typ: [file]})
    test.seek(0)

    comp_dict = loader(test)
    comp_dict = normalise(comp_dict, {dict: _to_complex})
    assert comp_dict == ref_dict

    # Check parse_single (i.e. wrapper) gives same result as specific loader
    comp_dict = parse_single(file, out_format=out_format)
    comp_dict = normalise(comp_dict, {dict: _to_complex})

    assert comp_dict == ref_dict


if __name__ == "__main__":
    pytest.main()



"""Parse castep .hug files."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO, TypedDict

from ..utilities.castep_res import labelled_floats
from ..utilities.utility import file_or_path, fix_data_types, stack_dict


class HugFileInfo(TypedDict):
    """Hugoniot information."""

    #: Percentage change in lattice parameters.
    compression: tuple[float, ...]
    #: Temperature at given compression.
    temperature: tuple[float, ...]
    #: Pressure at given compression.
    pressure: tuple[float, ...]
    #: Total energy at given compression.
    energy: tuple[float, ...]


@file_or_path(mode="r")
def parse_hug_file(hug_file: TextIO) -> HugFileInfo:
    """
    Parse castep .hug file.

    Parameters
    ----------
    hug_file
        Open handle to file to parse.

    Returns
    -------
    HugFileInfo
        Parsed info.
    """
    cols = ("compression", "temperature", "pressure", "energy")
    data = defaultdict(list)
    for line in hug_file:
        if match := re.search(labelled_floats(cols), line):
            stack_dict(data, match.groupdict())

    fix_data_types(data, dict.fromkeys(cols, float))
    return data



"""Parse castep .tddft files."""
from __future__ import annotations

import re
from typing import Literal, TextIO, TypedDict, Union

from ..utilities import castep_res as REs
from ..utilities.castep_res import get_numbers, labelled_floats
from ..utilities.datatypes import ComplexThreeVector, StandardHeader
from ..utilities.filewrapper import Block
from ..utilities.utility import file_or_path, to_type
from .parse_utilities import parse_regular_header

#: Overlap type
TDDFTOverlap = dict[Union[tuple[int, int], Literal["total"]], float]


class TDDFTSpectroData(TypedDict):
    """Spectroscopic data for single excitation."""

    #: Excitation character.
    characterisation: Literal["Singlet", "Doublet", "unknown"]
    #: Whether excitation is converged.
    converged: bool
    #: Dipole vectors.
    dipoles: ComplexThreeVector
    #: Energy of excitation.
    energy: float


class TDDFTFileInfo(StandardHeader, total=False):
    """TDDFT state occupation info."""

    #: Band overlap for excitations.
    overlap: list[TDDFTOverlap]
    #: Spectroscopic data of excitations.
    spectroscopic_data: list[TDDFTSpectroData]


@file_or_path(mode="r")
def parse_tddft_file(tddft_file: TextIO) -> TDDFTFileInfo:
    """
    Parse castep .tddft file.

    Parameters
    ----------
    tddft_file
        Open handle to file to parse.

    Returns
    -------
    TDDFTFileInfo
        Parsed info.
    """
    tddft_info = {}

    for line in tddft_file:
        if block := Block.from_re(line, tddft_file, "BEGIN header", "END header"):
            data = parse_regular_header(block, ("Higest occupied band",))
            tddft_info.update(data)

        elif block := Block.from_re(line, tddft_file,
                                    "BEGIN Characterisation of states as Kohn-Sham bands",
                                    "END Characterisation of states as Kohn-Sham bands"):

            tddft_info["overlap"] = []
            curr = {}

            for blk_line in block:
                if match := REs.TDDFT_STATE_RE.match(blk_line):
                    curr[int(match["occ"]), int(match["unocc"])] = float(match["overlap"])

                elif match := re.match(r"\s*Total overlap for state", blk_line):
                    curr["total"] = float(get_numbers(blk_line)[-1])
                    tddft_info["overlap"].append(curr)
                    curr = {}

        elif block := Block.from_re(line, tddft_file,
                                    "BEGIN TDDFT Spectroscopic Data",
                                    "END TDDFT Spectroscopic Data"):

            tddft_info["spectroscopic_data"] = []

            for blk_line in block:
                if match := re.match(rf"\s*\d+\s*{labelled_floats(('energy',))}"
                                     r"\s*(?P<characterisation>\w+)\s*"
                                     r"(?P<converged>Yes|No)\s*(?P<dipoles>.*)", blk_line):

                    match = match.groupdict()
                    match["converged"] = match["converged"] == "Yes"
                    match["energy"] = float(match["energy"])

                    dip = list(to_type(get_numbers(match["dipoles"]), float))
                    dip = list(map(complex, dip[0::2], dip[1::2]))

                    match["dipoles"] = dip
                    tddft_info["spectroscopic_data"].append(match)

    return tddft_info



"""Parse castep .efield files."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO

from ..utilities import castep_res as REs
from ..utilities.castep_res import labelled_floats
from ..utilities.constants import SND_D
from ..utilities.filewrapper import Block
from ..utilities.utility import file_or_path, fix_data_types, log_factory, stack_dict
from .parse_utilities import parse_regular_header


@file_or_path(mode="r")
def parse_efield_file(efield_file: TextIO) -> dict[str, float | int]:
    """
    Parse castep .efield files.

    Parameters
    ----------
    efield_file
        Open handle to file to parse.

    Returns
    -------
    dict[str, float | int]
        Parsed info.
    """
    # pylint: disable=too-many-branches,redefined-outer-name

    efield_info = defaultdict(list)
    logger = log_factory(efield_file)

    for line in efield_file:
        if block := Block.from_re(line, efield_file, "BEGIN header", "END header"):
            data = parse_regular_header(block, ("Oscillator Q",))
            efield_info.update(data)

        elif block := Block.from_re(line, efield_file, "BEGIN Oscillator Strengths",
                                    "END Oscillator Strengths"):

            logger("Found Oscillator Strengths")

            osc = defaultdict(list)
            block.remove_bounds(1, 2)
            for line in block:
                match = re.match(rf"\s*(?P<freq>{REs.INTNUMBER_RE})" +
                                 labelled_floats([*(f"S{d}" for d in SND_D)]), line)
                stack_dict(osc, match.groupdict())

            if osc:
                fix_data_types(osc, {"freq": float,
                                     **{f"S{d}": float for d in SND_D}})
                efield_info["oscillator_strengths"].append(osc)

        elif block := Block.from_re(line, efield_file, "BEGIN permittivity", "END permittivity"):

            logger("Found permittivity")

            perm = defaultdict(list)
            block.remove_bounds(1, 2)
            for line in block:
                match = re.match(labelled_floats(["freq",
                                                  *(f"e_r_{d}" for d in SND_D)]), line)
                stack_dict(perm, match.groupdict())

            if perm:
                fix_data_types(perm, {"freq": float,
                                      **{f"e_r_{d}": float for d in SND_D}})
                efield_info["permittivity"].append(perm)

    return efield_info



"""List of parsers for binary file formats."""
from __future__ import annotations

from collections.abc import Callable

from .cst_esp_file_parser import parse_cst_esp_file

#: Dictionary of available parsers.
PARSERS: dict[str, Callable] = {
    "cst_esp": parse_cst_esp_file,
}

#: Names of parsers/parsable file extensions (without ``"."``).
CASTEP_OUTPUT_NAMES: tuple[str, ...] = tuple(PARSERS.keys())

#: Names of parsable file extensions.
CASTEP_FILE_FORMATS: tuple[str, ...] = tuple(f".{typ}" for typ in CASTEP_OUTPUT_NAMES)



"""Main castep parser access routines."""
from __future__ import annotations

import io
import logging
import sys
from collections import ChainMap
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any, TextIO

from ..bin_parsers import PARSERS as BIN_PARSERS
from ..parsers import PARSERS
from ..parsers.castep_file_parser import Filters
from ..utilities.constants import OutFormats
from ..utilities.dumpers import get_dumpers
from ..utilities.utility import flatten_dict, json_safe, normalise
from .args import extract_parsables, parse_args

ALL_PARSERS = ChainMap(PARSERS, BIN_PARSERS)


def parse_single(in_file: str | Path | TextIO,
                 parser: Callable[[TextIO], list[dict[str, Any]]] | None = None,
                 out_format: OutFormats = "print",
                 *,
                 loglevel: int = logging.WARNING,
                 testing: bool = False) -> list[dict]:
    """
    Parse a file using the given parser and post-process according to options.

    Parameters
    ----------
    in_file
        Input file to parse.
    parser
        Castep parser to use. If `None` will be determined from extension.
    out_format
        Format to dump as.
    loglevel
        Logging level.
    testing
        Whether used for test suite (disable processing fragile properties).

    Returns
    -------
    dict
        Parsed data.

    Raises
    ------
    ValueError
        If invalid `parser` provided.
    """
    logging.basicConfig(format="%(levelname)s: %(message)s", level=loglevel)

    if isinstance(in_file, str):
        in_file = Path(in_file)

    if parser is None and isinstance(in_file, Path):
        ext = in_file.suffix.strip(".")

        if ext not in ALL_PARSERS:
            raise ValueError(f"Parser for file {in_file} (assumed type: {ext}) not found")

        parser = ALL_PARSERS[ext]

    if parser is None:
        raise ValueError("Unable to determine parser. Please specify through arguments.")

    if parser is ALL_PARSERS["castep"] and testing:
        data = parser(in_file, filters=Filters.TESTING)
    else:
        data = parser(in_file)

    if out_format == "json" or testing:
        data = normalise(data, {dict: json_safe, complex: json_safe})
    elif out_format in {"yaml", "ruamel"}:
        data = normalise(data, {tuple: list})

    if testing:
        if isinstance(data, Sequence):
            data = [flatten_dict(run) for run in data]
        else:
            data = flatten_dict(data)

    return data


def parse_all(
        output: str | Path | TextIO | None = None,
        out_format: OutFormats = "json",
        *,
        loglevel: int = logging.WARNING,
        testing: bool = False,
        **files: Sequence[Path],
) -> None:
    """
    Parse all files in files dict.

    Parameters
    ----------
    output
        Filepath or handle to dump output to.
    out_format
        Format to dump as.
    loglevel
        Logging level.
    testing
        Whether used for test suite (disable processing fragile properties).
    **files
        Dictionary of {parser needed: Sequence of paths to parse}.
    """
    file_dumper = get_dumpers(out_format)

    data = {}
    for typ, paths in files.items():
        parser = ALL_PARSERS[typ]
        for path in paths:
            data[path] = parse_single(path, parser, out_format, loglevel=loglevel, testing=testing)

    if len(data) == 1:
        data = data.popitem()[1]

    if output is None:
        file_dumper(data, sys.stdout)
    elif isinstance(output, io.TextIOBase):
        file_dumper(data, output)
    else:
        output = Path(output)
        with output.open("a+", encoding="utf-8") as out_file:
            file_dumper(data, out_file)


def main() -> None:
    """Run the main program from command line."""
    args = parse_args()
    dict_args = extract_parsables(args)

    parse_all(output=args.output,
              loglevel=getattr(logging, args.log.upper()),
              testing=args.testing,
              out_format=args.out_format,
              **dict_args)


if __name__ == "__main__":
    main()



"""Module for constants used in castep_outputs."""

from typing import Literal

#: Atomic orbitals.
SHELLS = ("s", "p", "d", "f")

#: First order cartesian directions.
FST_D = ("x", "y", "z")

#: Second order cartesian directions in Voigt notation.
SND_D = ("xx", "yy", "zz", "yz", "xz", "xy")

#: Valid CASTEP minimisers.
MINIMISERS = ("bfgs", "lbfgs", "fire", "tpsd", "dmd", "di")

#: Valid CASTEP pair potentials.
PAIR_POTS = ("LJ", "BUCK", "COUL", "DZ", "POL", "BB", "SHO",
             "SW", "MORS", "POLM", "LJ_S", "PES", "BU_S", "TIP4", "QUIP")

#: Aliases in .geom/.md files to fill names.
TAG_ALIASES = {"E": "energy",
               "T": "temperature",
               "P": "pressure",
               "h": "lattice_vectors",
               "hv": "lattice_velocity",
               "S": "stress",
               "R": "position",
               "V": "velocity",
               "F": "force"}

#: Input/output types used in transition state searches.
TS_TYPES = {"REA": "reagent",
            "PRO": "product",
            "TST": "test"}

#: Valid formats castep_outputs can dump to.
OutFormats = Literal["json", "pprint", "print", "ruamel", "yaml"]



"""Parse castep .err files."""
from __future__ import annotations

from typing import TextIO, TypedDict

from ..utilities.utility import file_or_path


class ErrFileInfo(TypedDict):
    """Error file information."""

    #: Message detailing cause of failure.
    message: str
    #: Backtrace at point of failure.
    stack: list[str]


@file_or_path(mode="r")
def parse_err_file(err_file: TextIO) -> ErrFileInfo:
    """
    Parse castep .err files.

    Parameters
    ----------
    err_file
        Open handle to file to parse.

    Returns
    -------
    ErrFileInfo
        Parsed info.
    """
    accum: ErrFileInfo = {"message": "", "stack": []}

    while "Current trace stack" not in (line := err_file.readline()):
        accum["message"] += line
    accum["message"] = accum["message"].strip()

    for line in err_file:
        accum["stack"].append(line.strip())

    return accum



"""Parse castep .phonon_dos files."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO, TypedDict

from ..utilities import castep_res as REs
from ..utilities.castep_res import labelled_floats
from ..utilities.datatypes import StandardHeader, ThreeVector
from ..utilities.filewrapper import Block
from ..utilities.utility import file_or_path, fix_data_types, log_factory, stack_dict
from .parse_utilities import parse_regular_header


class GradientInfo(TypedDict, total=False):
    """Info on phonon DOS gradients."""

    #: :math:`q`-point weight.
    pth: list[float]
    #: :math:`q`-point.
    qpt: ThreeVector
    #: Branch index.
    n: list[int]
    #: :math:`f` at point.
    f: list[float]
    #: :math:`\nabla_{q} \cdot f` at :math:`q`-point.
    Grad_qf: list[float]


class DOSInfo(TypedDict, total=False):
    """
    Density of states contribution.

    Notes
    -----
    Also includes per-species contributions whose keys are
    the species.
    """

    #: Frequencies.
    freq: list[float]
    #: Density out to Frequncy.
    g: list[float]


class PhononDosFileInfo(StandardHeader, total=False):
    """Phonon Density of states info."""

    #: Number of ions in system.
    ions: int
    #: Number of species in system.
    species: int
    #: Number of phonon branches.
    branches: int
    #: Phonon density of states information.
    dos: list[DOSInfo]
    #: Phonon DoS gradients information.
    gradients: list[GradientInfo]


@file_or_path(mode="r")
def parse_phonon_dos_file(phonon_dos_file: TextIO) -> PhononDosFileInfo:
    """
    Parse castep .phonon_dos file.

    Parameters
    ----------
    phonon_dos_file
        Open handle to file to parse.

    Returns
    -------
    PhononDosFileInfo
        Parsed info.
    """
    # pylint: disable=too-many-branches,redefined-outer-name
    logger = log_factory(phonon_dos_file)
    phonon_dos_info = defaultdict(list)

    for line in phonon_dos_file:
        if block := Block.from_re(line, phonon_dos_file, "BEGIN header", "END header"):
            data = parse_regular_header(block)
            phonon_dos_info.update(data)

        elif block := Block.from_re(line, phonon_dos_file, "BEGIN GRADIENTS", "END GRADIENTS"):

            logger("Found gradient block")
            qdata = defaultdict(list)

            def fix(qdat: dict) -> None:
                fix_data_types(qdat, {"qpt": float,
                                      "pth": float,
                                      "n": int,
                                      "f": float,
                                      "Grad_qf": float})

            for line in block:
                if match := REs.PHONON_PHONON_RE.match(line):
                    if qdata:
                        fix(qdata)
                        phonon_dos_info["gradients"].append(qdata)

                    qdata = defaultdict(list)

                    for key, val in match.groupdict().items():
                        qdata[key] = val.split()

                elif match := REs.PROCESS_PHONON_PHONON_RE.match(line):
                    stack_dict(qdata, match.groupdict())

            if qdata:
                fix(qdata)
                phonon_dos_info["gradients"].append(qdata)

        elif block := Block.from_re(line, phonon_dos_file, "BEGIN DOS", "END DOS"):

            logger("Found DOS block")

            dos = defaultdict(list)
            # First chunk is " BEGIN DOS   Freq (cm-1)  g(f)", thus need the 5th on
            species = block[0].split()[5:]
            headers = ("freq", "g", *species)
            rows = re.compile(labelled_floats(headers))

            block.remove_bounds(1, 2)

            for line in block:
                match = rows.match(line)
                stack_dict(dos, match.groupdict())

            if dos:
                fix_data_types(dos, dict.fromkeys(headers, float))
                phonon_dos_info["dos"].append(dos)

    return phonon_dos_info



"""Convenient filewrapper class."""
from __future__ import annotations

import re
from collections.abc import Sequence
from io import StringIO
from typing import NoReturn, TextIO, TypeVar

from .castep_res import Pattern

FWSelf = TypeVar("FWSelf", bound="FileWrapper")


class FileWrapper:
    """
    Convenience file wrapper to add rewind and line number capabilities.

    Parameters
    ----------
    file
        File to wrap and control.
    """

    def __init__(self, file: TextIO) -> FWSelf:
        self._file = file
        self._pos = 0
        self._lineno = 0

    def __iter__(self) -> FWSelf:
        return self

    def __next__(self) -> str:
        self._lineno += 1
        self._pos = self.file.tell()
        nextline = self.file.readline()
        if not nextline:
            raise StopIteration
        return nextline

    def rewind(self) -> None:
        """
        Rewind file to previous line.

        If iterated by `next` (or `for`) can rewind
        one line.

        Useful for blocks not terminated by a clear
        statement, where we can only check if next line
        does *NOT* match.
        """
        if self.file.tell() == self._pos:
            return
        self._lineno -= 1
        self.file.seek(self._pos)

    @property
    def file(self) -> TextIO:
        """
        File handle wrapped by the filewrapper.

        Returns
        -------
        ~typing.TextIO
            Underlying file.
        """
        return self._file

    @property
    def lineno(self) -> int:
        """
        Current line number.

        Returns
        -------
        int
            Current line number in file.
        """
        return self._lineno

    @property
    def name(self) -> str:
        """
        Name of underlying file.

        Returns
        -------
        str
            Name if name is known otherwise "unknown".
        """
        return self.file.name if hasattr(self.file, "name") else "unknown"


BSelf = TypeVar("BSelf", bound="Block")


class Block:
    """
    Data block class returned from :func:`get_block`.

    Emulates the properties of both a file, and sequence.
    """

    def __init__(self, parent: TextIO | FileWrapper | Block | None) -> Block:
        if isinstance(parent, (FileWrapper, Block)):
            self._lineno = parent.lineno
        else:
            self._lineno = 0

        self._name = parent.name if hasattr(parent, "name") else "unknown"
        self._i = -1
        self._data: tuple[str, ...] = ()

    @classmethod
    def get_lines(
            cls,
            in_file: TextIO | FileWrapper | Block,
            n_lines: int,
            *,
            eof_possible: bool = False,
    ) -> Block:
        r"""
        Read the next `n_lines` from `in_file` and return the block.

        Parameters
        ----------
        in_file
            File to read data from.
        n_lines
            Number of lines to read.

        Returns
        -------
        Block
            Read data.

        Raises
        ------
        OSError
            If EOF reached and ``not eof_possible``.

        Examples
        --------
        >>> from io import StringIO
        >>> x = StringIO('Hello\nThere\nFriend')
        >>> block = Block.get_lines(x, 2)
        >>> type(block).__name__
        'Block'
        >>> block.aslist()
        ['Hello\n', 'There\n']
        """
        block = cls(in_file)

        data: list[str] = []
        for i, line in enumerate(in_file, 1):
            if i > n_lines:
                break
            data.append(line)
        else:
            if not eof_possible:
                if hasattr(in_file, "name"):
                    raise OSError(f"Unexpected end of file in {in_file.name}.")
                raise OSError("Unexpected end of file.")

        block._data = tuple(data)
        return block

    @classmethod
    def from_re(
            cls,
            init_line: str,
            in_file: TextIO | FileWrapper | Block,
            start: Pattern,
            end: Pattern,
            *,
            n_end: int = 1,
            eof_possible: bool = False,
    ) -> Block:
        """
        Check if line is the start of a block and return the block if it is.

        Parameters
        ----------
        init_line
            Initial line which may start the block.
        in_file
            File handle to read data from.
        start
            RegEx matched against `init_line` to see if is start of block.
        end
            RegEx to verify if block has ended.
        n_end
            Number of times `end` must match before block is returned.
        eof_possible
            Whether it is possible block is ended by EOF.

        Returns
        -------
        Block
            Recovered block of data matching prereqs.

        Notes
        -----
        Advances `in_file` as it does so.

        Raises
        ------
        OSError
            If EOF reached and ``not eof_possible``.
        """
        block = cls(in_file)

        if not re.search(start, init_line):
            # Return empty block. (bool -> False)
            return block

        data: list[str] = []
        data.append(init_line)

        found = 0
        for line in in_file:
            data.append(line)
            if re.search(end, line):
                found += 1
                if found == n_end:
                    break
        else:
            if not eof_possible:
                if hasattr(in_file, "name"):
                    raise OSError(f"Unexpected end of file in {in_file.name}.")
                raise OSError("Unexpected end of file.")

        block._data = tuple(data)
        return block

    @classmethod
    def from_iterable(
            cls,
            data: Sequence[str],
            parent: TextIO | FileWrapper | Block | None = None,
    ) -> BSelf:
        r"""
        Construct a Block from an iterable containing strings.

        Parameters
        ----------
        data
            Data to read into block.

        Returns
        -------
        Self
            Block from an iterable.

        Examples
        --------
        >>> x = Block.from_iterable(("Hello", "There"))
        >>> type(x).__name__
        'Block'
        >>> str(x)
        'Hello\nThere'
        """
        block = cls(parent)
        block._data = tuple(data)
        return block

    def remove_bounds(self, /, fore: int = 1, back: int = 2) -> None:
        """
        Remove the bounding lines of the block.

        These are often region delimiters and not
        relevant to the processed data.

        Parameters
        ----------
        fore
            Whether to strip leading line from data.
        back
            Whether to strip trailing line from data.
        """
        self._lineno += fore
        self._data = self._data[fore:
                                -back if back else None]

    def asstringio(self) -> StringIO:
        """
        Return block as a simulated file.

        Returns
        -------
        ~io.StringIO
            Block data as StringIO
        """
        return StringIO(str(self))

    def aslist(self) -> list[str]:
        """
        Return block as a list of lines.

        Returns
        -------
        list[str]
            Block as list of lines.
        """
        return list(self._data)

    def rewind(self) -> None:
        """
        Rewind to previous line.

        Useful for blocks not terminated by a clear
        statement, where we can only check if next line
        does *NOT* match.

        Raises
        ------
        ValueError
            Block already rewound fully.
        """
        if self._i < 0:
            self._i = -1
            raise ValueError("Block already at beginning.")

        self._i -= 1

    def __bool__(self) -> bool:
        return any(map(str.strip, self._data))

    def __str__(self) -> str:
        return "\n".join(self._data)

    def __iter__(self) -> BSelf:
        return self

    def __next__(self) -> str:
        self._i += 1
        if self._i >= len(self):
            raise StopIteration
        return self._data[self._i]

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: int | slice) -> str:
        return self._data[key]

    @property
    def name(self) -> str:
        """
        Name of underlying file.

        Returns
        -------
        str
            Name if name is known otherwise "unknown".
        """
        return self._name

    @property
    def lineno(self) -> int:
        """
        Current line number.

        Returns
        -------
        int
            Current line number in file.
        """
        return self._lineno + self._i

    @property
    def file(self) -> NoReturn:
        """
        Block has no internal file.

        Raises
        ------
        NotImplementedError
            Block has no internal file holder.
        """
        raise NotImplementedError("Block has no internal file holder.")
